"""
Module containing the corrections for estimating and correcting, and subtracting
the contribution of the sky emission in DataContainers.
"""
# =============================================================================
# Basics packages
# =============================================================================
import numpy as np
import copy
import os
from time import time
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm, Normalize
import scipy
# =============================================================================
# Astropy and associated packages
# =============================================================================
from astropy.io import fits
from astropy.modeling import models, fitting
from astropy.table import QTable
from astropy import stats
from astropy import units as u
import scipy.ndimage
import scipy.signal

# =============================================================================
# PyKOALA
# =============================================================================
from pykoala import vprint
from pykoala.plotting.utils import new_figure, plot_image, plot_fibres
from pykoala.corrections.correction import CorrectionBase
from pykoala.corrections.throughput import Throughput
from pykoala.corrections.wavelength import WavelengthOffset
from pykoala.data_container import DataContainer, RSS
from pykoala.ancillary import check_unit, symmetric_background


#TODO: Move to a math module
class BackgroundEstimator(object):
    """
    Class for estimating background and its dispersion using different statistical methods.
    """

    @staticmethod
    def mean(rss_intensity):
        """
        Compute the mean and standard deviation of the spectra.

        Parameters
        ----------
        rss_intensity : np.ndarray
            Input array (n_fibres x n_wavelength).

        Returns
        -------
        background : np.ndarray
            The computed background (mean) of the data.
        background_sigma : np.ndarray
            The dispersion (standard deviation) of the data.
        """
        background = np.nanmean(rss_intensity, axis=0)
        background_sigma = np.nanstd(rss_intensity, axis=0)
        return background, background_sigma

    @staticmethod
    def percentile(rss_intensity, percentiles=[16, 50, 84]):
        """
        Compute the background and dispersion from specified percentiles.

        Parameters
        ----------
        rss_intensity : np.ndarray
            Input array (n_fibres x n_wavelength).
        percentiles : list of float, optional
            The percentiles to use for computation. Default is [16, 50, 84].

        Returns
        -------
        background : np.ndarray
            The computed background (median) of the data.
        background_sigma : np.ndarray
            The dispersion (half the interpercentile range) of the data.
        """
        plow, background, pup = np.nanpercentile(
            rss_intensity, percentiles, axis=0)
        background_sigma = (pup - plow) / 2
        return background, background_sigma

    @staticmethod
    def mad(rss_intensity):
        """
        Estimate the background from the median and the dispersion from
        the Median Absolute Deviation (MAD) along the given axis.

        Parameters
        ----------
        rss_intensity : np.ndarray
            Input array (n_fibres x n_wavelength).

        Returns
        -------
        background : np.ndarray
            The computed background (median) of the data.
        background_sigma : np.ndarray
            The dispersion (scaled MAD) of the data.
        """
        background = np.nanmedian(rss_intensity, axis=0)
        mad = np.nanmedian(np.abs(rss_intensity - background[np.newaxis:]),
                           axis=0)
        background_sigma = 1.4826 * mad
        return background, background_sigma

    @staticmethod
    def biweight(rss_intensity):
        """
        Estimate the background and dispersion using the biweight method.

        Parameters
        ----------
        rss_intensity : np.ndarray
            Input array (n_fibres x n_wavelength).

        Returns
        -------
        background : np.ndarray
            Computed background (location) of the data.
        background_sigma : np.ndarray
            Dispersion (scale) of the data.
        """
        background = stats.biweight_location(rss_intensity, axis=0)
        background_sigma = stats.biweight_scale(rss_intensity, axis=0)
        return background, background_sigma

    @staticmethod
    def mode(rss_intensity):
        """
        Estimate the background and dispersion using the mode.

        Parameters
        ----------
        rss_intensity : np.ndarray
            Input array (n_fibres x n_wavelength).

        Returns
        -------
        background : np.ndarray
            The computed background sky.
        background_sigma : np.ndarray
            Scaled MAD of the data behind the mode.
        """
        mode = np.empty_like(rss_intensity[0])
        sigma = np.empty_like(mode)
        for wavelength_index in range(mode.size):
            fibre_flux = rss_intensity[:, wavelength_index]
            mode[wavelength_index], dummy = symmetric_background(fibre_flux)
            below_mode = fibre_flux < mode[wavelength_index]
            sigma[wavelength_index] = 1.4826 * np.nanmedian(
                mode[wavelength_index] - fibre_flux[below_mode])

        return mode, sigma

#TODO : Move to a math module
# =============================================================================
# Continuum estimators
# =============================================================================
class ContinuumEstimator:
    """
    Class for estimating the continuum of spectral data using different methods.

    Methods
    -------
    medfilt_continuum(data, window_size=5)
        Estimate the continuum using a median filter.

    pol_continuum(data, wavelength, pol_order=3, **polfit_kwargs)
        Estimate the continuum using polynomial fitting.
    """

    @staticmethod
    def medfilt_continuum(data, window_size=5):
        """
        Estimate the continuum using a median filter.

        Parameters
        ----------
        data : np.ndarray
            The input data array for which to compute the continuum.
        window_size : int, optional
            The size of the window over which to compute the median filter. Default is 5.

        Returns
        -------
        continuum : np.ndarray
            The estimated continuum of the input data.
        """
        continuum = scipy.signal.medfilt(data, window_size)
        return continuum

    @staticmethod
    def percentile_continuum(data, percentile, window_size=5):
        """
        Estimate the continuum using a percentile filter.

        Parameters
        ----------
        data : np.ndarray
            The input data array for which to compute the continuum.
        percentile : list or tuple
            The percentiles (0-100) to use for the continuum estimation.
        window_size : int, optional
            The size of the window over which to compute the percentile filter. Default is 5.

        Returns
        -------
        continuum : np.ndarray
            The estimated continuum of the input data.
        """
        continuum = scipy.ndimage.percentile_filter(data, percentile,
                                                    window_size)
        return continuum

    @staticmethod
    def pol_continuum(data, wavelength, pol_order=3, **polfit_kwargs):
        """
        Estimate the continuum using polynomial fitting.

        Parameters
        ----------
        data : np.ndarray
            The input data array for which to compute the continuum.
        wavelength : np.ndarray
            The wavelength array corresponding to the data.
        pol_order : int, optional
            The order of the polynomial to fit. Default is 3.
        **polfit_kwargs : dict, optional
            Additional keyword arguments to pass to `np.polyfit`.

        Returns
        -------
        continuum : np.ndarray
            The estimated continuum of the input data.
        """
        fit = np.polyfit(wavelength, data, pol_order, **polfit_kwargs)
        polynomial = np.poly1d(fit)
        return polynomial(wavelength)

    default_min_separation = 10

    @classmethod
    def lower_envelope(self, x, y, min_separation=None):
        '''
        #TODO --> Refactor
        Fit lower envelope of a single spectrum:
        1) Find local minima, with a minimum separation `min_separation`.
        2) Interpolate linearly between them.
        3) Add "typical" (~median) offset.
        '''
        if min_separation is None:
            min_separation = self.default_min_separation
        valleys = []
        y[np.isnan(y)] = np.inf
        for i in range(min_separation, y.size-min_separation-1):
            if np.argmin(y[i-min_separation:i+min_separation+1]) == min_separation:
                valleys.append(i)
        y[~np.isfinite(y)] = np.nan

        continuum = np.fmin(y, np.interp(x, x[valleys], y[valleys]))

        offset = y - continuum
        offset = np.nanpercentile(offset[offset > 0], np.linspace(1, 50, 51))
        density = (np.arange(offset.size) + 1) / offset
        offset = np.median(offset[density > np.max(density)/2])

        return continuum+offset, offset

#TODO: Documentation
class ContinuumModel:

    def __init__(self, dc: DataContainer, min_separation=None):
        self.update(dc, min_separation)

    # TODO: Don't assume RSS format (intensity[spec_id, wavelength])
    #       def update(self, dc:DataContainer, min_separation=None):
    def update(self, dc: RSS, min_separation=None):
        n_spectra = dc.intensity.shape[0]
        self.intensity = np.zeros_like(dc.intensity)
        self.scale = np.zeros(n_spectra)

        print(f"> Find continuum for {n_spectra} spectra:")
        t0 = time()

        for i in range(n_spectra):
            self.intensity[i], self.scale[i] = ContinuumEstimator.lower_envelope(
                dc.wavelength, dc.intensity[i], min_separation)

        print(f"  Done ({time()-t0:.3g} s)")
        self.strong_sky_lines = self.detect_lines(dc)

    def detect_lines(self, dc: DataContainer, n_sigmas=3):
        SN = (dc.intensity - self.intensity) / self.scale[:, np.newaxis]
        SN_p16, SN_p50, SN_p84 = np.nanpercentile(SN, [16, 50, 84], axis=0)

        line_mask = SN_p16 - (n_sigmas-1)*(SN_p84-SN_p16)/2 > 0
        line_mask[0] = False
        line_mask[-1] = False

        line_left = np.where(~line_mask[:-1] & line_mask[1:])[0]
        line_right = np.where(line_mask[:-1] & ~line_mask[1:])[0]
        line_right += 1
        print(f'  {line_left.size} strong sky lines ({np.count_nonzero(line_mask)} out of {dc.wavelength.size} wavelengths)')
        return QTable((line_left, line_right), names=('left', 'right'))


# =============================================================================
# Line Spread Function
# =============================================================================
#TODO: Documenation and move to spectra module
class LSF_estimator:

    def __init__(self, wavelength_range, resolution):
        self.delta_lambda = np.linspace(-wavelength_range, wavelength_range, int(
            np.ceil(2 * wavelength_range / resolution)) + 1)

    def find_LSF(self, line_wavelengths, wavelength, intensity):
        line_spectra = np.zeros(
            (len(line_wavelengths), self.delta_lambda.size))
        for i, line_wavelength in enumerate(line_wavelengths):
            sed = np.interp(line_wavelength + self.delta_lambda,
                            wavelength, intensity)
            line_spectra[i] = self.normalise(sed)
        return self.normalise(np.nanmedian(line_spectra, axis=0))

    def normalise(self, x):
        x -= stats.biweight.biweight_location(x)  # subtract continuum
        # x -= np.median(x)
        norm = x[x.size//2]  # normalise at centre
        if norm > 0:
            x /= norm
        else:
            x *= np.nan
        return x

    def find_FWHM(self, profile):
        threshold = np.max(profile)/2
        left = np.max(
            self.delta_lambda[(self.delta_lambda < 0) & (profile < threshold)])
        right = np.min(
            self.delta_lambda[(self.delta_lambda > 0) & (profile < threshold)])
        return right-left


# =============================================================================
# Sky lines library
# =============================================================================
# TODO : this should be an instance of EMission lines in the spectra module
def uves_sky_lines():
    """
    Library of sky emission lines measured with UVES@VLT.

    For more details, see the `UVES Sky Spectrum <https://www.eso.org/observing/dfo/quality/UVES/pipeline/sky_spectrum.html>`_.

    Returns
    -------
    line_wavelength : np.ndarray
        Array containing the wavelength positions of each emission line centroid in Angstroms.
    line_fwhm : np.ndarray
        Array containing the FWHM values of each line expressed in Angstroms.
    line_flux : np.ndarray
        Array containing the flux of each line expressed in 1e-16 ergs/s/A/cm^2/arcsec^2.
    """
    # Prefix of each table
    prefix = ["346", "437", "580L", "580U", "800U", "860L", "860U"]
    # Path to tables
    data_path = os.path.join(
        os.path.dirname(__file__), "..", "input_data", "sky_lines", "ESO-UVES")
    if not os.path.isdir(data_path):
        raise FileNotFoundError(f'Directory {data_path} not in {os.listdir(os.path.join(os.path.dirname(__file__), "..", "input_data", "sky_lines"))}')
    
    # Initialize arrays to store line properties
    line_wavelength = np.empty(0)
    line_fwhm = np.empty(0)
    line_flux = np.empty(0)

    # Read data from each file
    for p in prefix:
        file = os.path.join(data_path, f"gident_{p}.tfits")
        if not os.path.isfile(file):
            raise FileNotFoundError(f"File '{file}' could not be found in {os.listdir(data_path)}")

        with fits.open(file) as f:
            wave = f[1].data['LAMBDA_AIR']
            fwhm = f[1].data['FWHM']
            flux = f[1].data['FLUX']

            line_wavelength = np.hstack((line_wavelength, wave))
            line_fwhm = np.hstack((line_fwhm, fwhm))
            line_flux = np.hstack((line_flux, flux))

    line_fwhm = line_fwhm << u.angstrom
    line_wavelength = line_wavelength << u.angstrom
    # Sort lines by wavelength
    sort_pos = np.argsort(line_wavelength)
    return line_wavelength[sort_pos], line_fwhm[sort_pos], line_flux[sort_pos]

# =============================================================================
# Sky models
# =============================================================================


class SkyModel(object):
    """
    Abstract class for a sky emission model.

    Attributes
    ----------
    wavelength : np.ndarray
        1-D array representing the wavelengths of the sky model.
    intensity : np.ndarray
        Array representing the intensity of the sky model.
    variance : np.ndarray
        Array representing the variance associated with the sky model.
        It must have the same dimensions as `intensity`.
    continuum : np.ndarray
        Array representing the continuuum emission of the sky model.
    sky_lines: np.ndarray
        1-D array representing the wavelength of a collection of sky emission
        lines expressed in angstrom. This is used for fitting the emission lines
        from the (continuum-substracted) intensity.
    verbose : bool, optional
        If True, print messages during execution. Default is True.

    Methods
    -------
    subtract(data, variance, axis=-1)
        Subtracts the sky model from the given data.

    subtract_pca()
        Placeholder for PCA subtraction method.

    vprint(*messages)
        Print messages if `verbose` is True.
    """

    verbose = True
    sky_lines = None

    def __init__(self, **kwargs):
        """
        Initialize the SkyModel object.

        Parameters
        ----------
        **kwargs : dict, optional
            Dictionary of parameters to initialize the SkyModel.
            Accepted keys are:

            - wavelength : np.ndarray
                1-D array representing the wavelengths of the sky model.
            - intensity : np.ndarray
                Array representing the intensity of the sky model.
            - variance : np.ndarray
                Array representing the variance associated with the sky model.
            - verbose : bool
                If True, print messages during execution. Default is True.
        """
        self.wavelength = check_unit(kwargs.get('wavelength', None), u.angstrom)
        self.intensity = check_unit(kwargs.get('intensity', None))
        self.variance = check_unit(kwargs.get('variance', None))
        self.continuum = check_unit(kwargs.get('continuum', None))

    def substract(self, data, variance, axis=-1, verbose=False):
        """
        Subtracts the sky model from the given data.

        Parameters
        ----------
        data : np.ndarray
            Data array from which the sky will be subtracted.
        variance : np.ndarray
            Array of variance data to include errors in determining the sky.
        axis : int, optional
            Spectral direction of the data. Default is -1.

        Returns
        -------
        data_subs : np.ndarray
            Data array after the sky model has been subtracted.
        var_subs : np.ndarray
            Variance array after including the sky model variance.
        """
        if data.ndim == 3 and self.intensity.ndim == 1:
            skymodel_intensity = self.intensity[:, np.newaxis, np.newaxis]
            skymodel_var = self.variance[:, np.newaxis, np.newaxis]
        elif data.ndim == 2 and self.intensity.ndim == 1:
            skymodel_intensity = self.intensity[np.newaxis, :]
            skymodel_var = self.variance[np.newaxis, :]
        elif data.ndim == 2 and self.intensity.ndim == 2:
            skymodel_intensity = self.intensity
            skymodel_var = self.variance
        else:
            vprint(
                f"Data dimensions ({data.shape}) cannot be reconciled with "
                + f"sky mode ({self.intensity.shape})")
        data_subs = data - skymodel_intensity
        var_subs = variance + skymodel_var
        return data_subs, var_subs

    def substract_pca():
        """
        Placeholder for PCA subtraction method.

        This method is not yet implemented.
        """
        pass

    def remove_continuum(self, cont_estimator="median", cont_estimator_args=None):
        """
        Remove the continuum from the background model.

        Parameters
        ----------
        cont_estimator : str, optional
            Method to estimate the continuum signal. Default is 'median'.
        cont_estimator_args : dict, optional
            Arguments for the continuum estimator. Default is None.
        """
        if cont_estimator_args is None:
            cont_estimator_args = {}
        if self.intensity is not None:
            if hasattr(ContinuumEstimator, cont_estimator):
                estimator = getattr(ContinuumEstimator, cont_estimator)
            else:
                raise NameError(f"{cont_estimator} does not correspond to any"
                                + "available continuum method")
            self.continuum = estimator(self.intensity, **cont_estimator_args)
            self.intensity -= self.continuum
        else:
            raise AttributeError("Sky model intensity has not been computed")

    # TODO: This should make use of the methods in the "spectra" module
    def fit_emission_lines(self, window_size=100,
                           resampling_wave=0.1, **fit_kwargs):
        """
        Fit emission lines to the continuum-subtracted spectrum.

        Parameters
        ----------
        window_size : int, optional
            Size of the wavelength window for fitting. Default is 100.
        resampling_wave : float, optional
            Wavelength resampling interval. Default is 0.1.

        Returns
        -------
        emission_model : models.Gaussian1D
            Fitted emission line model.
        emission_spectra : np.ndarray
            Emission spectra.
        """
        assert self.intensity is not None, "Sky Model intensity is None"

        if self.continuum is None:
            vprint("Sky Model intensity might contain continuum emission"
                   " leading to unsuccessful emission line fit")
        if self.variance is None:
            errors = np.ones_like(self.intensity, dtype=float)

        finite_mask = np.isfinite(self.intensity)
        p0_amplitude = np.interp(self.sky_lines, self.wavelength[finite_mask],
                                 self.intensity[finite_mask])
        p0_amplitude = np.clip(p0_amplitude, a_min=0, a_max=None)
        fit_g = fitting.LevMarLSQFitter()
        emission_model = models.Gaussian1D(amplitude=0, mean=0, stddev=0)
        emission_spectra = np.zeros_like(self.wavelength)
        wavelength_windows = np.arange(self.wavelength.min(),
                                       self.wavelength.max(), window_size)
        wavelength_windows[-1] = self.wavelength.max()
        vprint(f"Fitting all emission lines ({self.sky_lines.size})"
                    + " to continuum-subtracted sky spectra")
        for wl_min, wl_max in zip(wavelength_windows[:-1], wavelength_windows[1:]):
            vprint(f"Starting fit in the wavelength range [{wl_min:.1f}, "
                   + f"{wl_max:.1f}]")
            mask_lines = (self.sky_lines >= wl_min) & (self.sky_lines < wl_max)
            mask = (self.wavelength >= wl_min) & (
                self.wavelength < wl_max) & finite_mask
            wave = np.arange(self.wavelength[mask][0],
                             self.wavelength[mask][-1], resampling_wave)
            obs = np.interp(wave, self.wavelength[mask], self.intensity[mask])
            err = np.interp(wave, self.wavelength[mask], errors[mask])
            if mask_lines.any():
                vprint(f"> Line to Fit {self.sky_lines[mask_lines][0]:.1f}")
                window_model = models.Gaussian1D(
                    amplitude=p0_amplitude[mask_lines][0],
                    mean=self.sky_lines[mask_lines][0],
                    stddev=1,
                    bounds={'amplitude': (p0_amplitude[mask_lines][0] * 0.5, p0_amplitude[mask_lines][0] * 10),
                            'mean': (self.sky_lines[mask_lines][0] - 5, self.sky_lines[mask_lines][0] + 5),
                            'stddev': (self.sky_lines_fwhm[mask_lines][0] / 2, 5)}
                )
                for line, p0, sigma in zip(
                        self.sky_lines[mask_lines][1:], p0_amplitude[mask_lines][1:],
                        self.sky_lines_fwhm[mask_lines][1:]):
                    vprint(f"Line to Fit {line:.1f}")
                    model = models.Gaussian1D(
                        amplitude=p0, mean=line, stddev=sigma,
                        bounds={'amplitude': (p0 * 0.5, p0 * 10), 'mean': (line - 5, line + 5),
                                'stddev': (sigma / 2, 5)}
                    )
                    window_model += model
                g = fit_g(window_model, wave, obs,
                          weights=1 / err, **fit_kwargs)
                emission_spectra += g(self.wavelength)
                emission_model += g
        return emission_model, emission_spectra

    # TODO: This should create an instance of EmissionLines
    def load_sky_lines(self, path_to_table=None, lines_pct=84., **kwargs):
        """
        Load sky lines from a file.

        Parameters
        ----------
        path_to_table : str, optional
            Path to the table containing sky lines. Default is None.
        lines_pct : float, optional
            Percentile for selecting faint lines. Default is 84.
        kwargs : dict, optional
            Additional arguments for `np.loadtxt`.

        Returns
        -------

        """
        if path_to_table is not None:
            vprint(f"Loading input sky line table {path_to_table}")
            path_to_table = os.path.join(os.path.dirname(__file__),
                                         'input_data', 'sky_lines',
                                         path_to_table)
            self.sky_lines, self.sky_lines_fwhm, self.sky_lines_f = np.loadtxt(
                path_to_table, usecols=(0, 1, 2), unpack=True, **kwargs)
        else:
            vprint("Loading UVES sky line table")
            self.sky_lines, self.sky_lines_fwhm, self.sky_lines_f = uves_sky_lines()
        # Select only those lines within the wavelength range of the model
        common_lines = (self.sky_lines >= self.wavelength[0]) & (
            self.sky_lines <= self.wavelength[-1])
        self.sky_lines = self.sky_lines[common_lines]
        self.sky_lines_fwhm = self.sky_lines_fwhm[common_lines]
        self.sky_lines_f = self.sky_lines_f[common_lines]
        vprint(f"Total number of sky lines: {self.sky_lines.size}")
        # Blend sky emission lines
        delta_lambda = self.wavelength[1] - self.wavelength[0]
        vprint("Blending sky emission lines according to"
                + f"wavelength resolution ({delta_lambda} AA)")
        unresolved_lines = np.where(np.diff(self.sky_lines) <= delta_lambda)[0]
        while len(unresolved_lines) > 0:
            self.sky_lines[unresolved_lines] = (
                self.sky_lines[unresolved_lines] + self.sky_lines[unresolved_lines + 1]) / 2
            self.sky_lines_fwhm[unresolved_lines] = np.sqrt(
                self.sky_lines_fwhm[unresolved_lines]**2 +
                self.sky_lines_fwhm[unresolved_lines + 1]**2
            )
            self.sky_lines_f[unresolved_lines] = (
                self.sky_lines_f[unresolved_lines] +
                self.sky_lines_f[unresolved_lines + 1]
            )
            self.sky_lines = np.delete(self.sky_lines, unresolved_lines)
            self.sky_lines_fwhm = np.delete(
                self.sky_lines_fwhm, unresolved_lines)
            self.sky_lines_f = np.delete(self.sky_lines_f, unresolved_lines)
            unresolved_lines = np.where(
                np.diff(self.sky_lines) <= delta_lambda)[0]
        vprint(f"Total number of sky lines after blending: {self.sky_lines.size}")
        # Remove faint lines
        # self.vprint(f"Selecting the  sky lines after blending: {self.sky_lines.size}")
        faint = np.where(self.sky_lines_f < np.nanpercentile(
            self.sky_lines_f, lines_pct))[0]
        self.sky_lines = np.delete(self.sky_lines, faint)
        self.sky_lines_fwhm = np.delete(self.sky_lines_fwhm, faint)
        self.sky_lines_f = np.delete(self.sky_lines_f, faint)

    def plot_sky_model(self, show=False, fig_name='sky_model'):
        """Plot the sky model

        Parameters
        ----------
        show : bool
            Show the resulting plot. Default is False.

        Returns
        -------
        fig : :class:`matplotlib.pyplot.Figure`
            Figure containing the Sky Model plot.
        """
        if self.intensity is None:
            return None
        
        elif self.intensity.ndim == 1:
            fig, axes = new_figure(fig_name)
            if 'dc' in self.__dict__:
                title = f"1-D Sky Model for {self.dc.info['name']}"
            else:
                title = '1-D Sky Model'
            fig.suptitle(title)
            ax = axes[0, 0]
            ax.plot(self.wavelength, self.intensity,
                    color='b', alpha=0.5, label='sky intensity')
            if self.variance is not None:
                ax.fill_between(self.wavelength, self.intensity - self.variance**0.5,
                                self.intensity + self.variance**0.5, color='b',
                                alpha=0.2, label='uncertainty')
            ax.set_ylim(np.nanpercentile(self.intensity, [1, 99]).value)
            ax.set_ylabel(f"Intensity [{self.intensity.unit}]")
            ax.set_xlabel(f"Wavelength [{self.wavelength.unit}]")
            ax.legend()

        elif self.intensity.ndim == 2:
            fig = plt.figure(fig_name, constrained_layout=True)
            im_args = dict(
                interpolation='none', aspect='auto', origin="lower",
                extent=(1, self.intensity.shape[0],
                        self.wavelength[0], self.wavelength[-1]))
            ax = fig.add_subplot(121, title='2-D Sky Model Intensity')
            mappable = ax.imshow(self.intensity, **im_args)
            plt.colorbar(mappable, ax=ax)
            ax = fig.add_subplot(122, title='2-D Sky Model STD')
            mappable = ax.imshow(self.variance**0.5, **im_args)
            plt.colorbar(mappable, ax=ax)

        #if show:
        #    plt.show(fig_name)
        #else:
        if not show:
            plt.close(fig_name)
        return fig


class SkyOffset(SkyModel):
    """
    Sky model based on a single RSS offset sky exposure.

    This class builds a sky emission model from individual sky exposures.

    Attributes
    ----------
    dc : DataContainer
        Data container used to estimate the sky.
    exptime : float
        Net exposure time from the data container.
    """

    def __init__(self, dc):
        """
        Initialize the SkyOffset model with a data container.

        Parameters
        ----------
        dc : DataContainer
            Data container used to estimate the sky.
        """
        self.dc = dc
        self.exptime = dc.info['exptime']
        super().__init__()

    def estimate_sky(self):
        """
        Estimate the sky emission model.

        This method calculates the intensity and variance of the sky model using
        percentiles, then normalizes them by the exposure time.
        """
        #TODO: This should create a 2D model that uses the median sky
        # but shifts it to the posiotoin given on each individual fibre
        self.intensity, self.variance = BackgroundEstimator.percentile(
            self.dc.rss_intensity, percentiles=[16, 50, 84])
        self.intensity, self.variance = (
            self.intensity / self.exptime,
            self.variance / self.exptime)


class SkyFromObject(SkyModel):
    """
    Sky model based on a single Data Container.

    This class builds a sky emission model using the data
    from a given Data Container that includes the contribution
    of an additional source (i.e. star/galaxy).

    Attributes
    ----------
    dc : DataContainer
        Input DataContainer object.
    exptime : float
        Net exposure time from the data container.
    bckgr : np.ndarray or None
        Estimated background. Initialized as None.
    bckgr_sigma : np.ndarray or None
        Estimated background standard deviation. Initialized as None.
    continuum : np.ndarray or None
        Estimated continuum. Initialized as None.
    """

    def __init__(self, dc,
                 bckgr_estimator='mad', bckgr_params=None,
                 sky_fibres=None, source_mask_nsigma=None,
                 remove_cont=False,
                 cont_estimator='median', cont_estimator_args=None,
                 qc_plots={'show': True}):
        """
        Initialize the SkyFromObject model.

        Parameters
        ----------
        dc : DataContainer
            Input DataContainer object.
        bckgr_estimator : str, optional
            Background estimator method to be used. Default is 'mad'.
        bckgr_params : dict, optional
            Parameters for the background estimator. Default is None.
        sky_fibres : None, 'auto', 'all', or list of 1D indices (optional)
            Default is None, which will read the sky fibre list from
            the `info` attribute; if absent or empty, it will trigger 'auto'.
            The 'auto' option will call `estimate_sky_fibres` to identify
            sky fibres based on a mean intensity threshold, automatically
            determined from the shape of the normalised spectra.
            Selecting `all` will use every spectrum in the `DataContainer`.
            Othrwise, this argument will be interpreted as a list of
            integer indices identifying sky spectra, assuming 1D RSS order.
        source_mask_nsigma : float, optional
            Sigma level for masking sources. Default is None.
        remove_cont : bool, optional
            If True, the continuum will be removed. Default is False.
        cont_estimator : str, optional
            Method to estimate the continuum signal. Default is 'median'.
        cont_estimator_args : dict, optional
            Arguments for the continuum estimator. Default is None.
        qc_plots: dict
            Dictionary to control QC plots.
        """
        vprint("Creating SkyModel from input Data Container")
        self.dc = dc
        # self.exptime = dc.info['exptime']
        #vprint(f"Estimating sky background contribution from the {bckgr_estimator}...")
        
        self.qc_plots = {}

        bckg, bckg_sigma = self._estimate_background(
            bckgr_estimator, bckgr_params, sky_fibres, source_mask_nsigma, qc_plots)
        super().__init__(wavelength=self.dc.wavelength,
                         intensity=bckg,
                         variance=bckg_sigma**2)
        if remove_cont:
            vprint("Removing background continuum")
            self.remove_continuum(cont_estimator, cont_estimator_args)
        
        if len(qc_plots) > 0:
            show_plot = qc_plots.get('show', True)
            plot_filename = qc_plots.get('filename_base', None)
            fig_name = 'sky_model'
            fig = self.plot_sky_model(show=show_plot, fig_name=fig_name)
            ax = fig.axes[0]
            p16, p50, p84 = np.nanpercentile(self.dc.rss_intensity,
                                             [16, 50, 84], axis=0)
            ax.plot(self.dc.wavelength, p50, 'k-', alpha=.1)
            ax.fill_between(self.dc.wavelength, p16, p84,
                            color='k', alpha=.1,
                            label=f"{self.dc.info['name']} (all fibres)")
            if 'sky_CASU' in self.dc.info:
                ax.plot(self.dc.wavelength, self.dc.info['sky_CASU'],
                        'r-', alpha=.5, label='WEAVE pipeline')
            ax.legend()
            self.qc_plots[fig_name] = fig
            if plot_filename is not None:
                fig.savefig(f'{plot_filename}_{fig_name}.png')
            if not show_plot:
                plt.close(fig_name)

    def _estimate_background(self, bckgr_estimator, bckgr_params=None,
                            sky_fibres=None, source_mask_nsigma=None,
                            qc_plots={}):
        """
        Estimate the background.

        Parameters
        ----------
        bckgr_estimator : str
            Background estimator method. Available methods: 'mad', 'percentile'.
        bckgr_params : dict, optional
            Parameters for the background estimator. Default is None.
        sky_fibres : None, 'auto', 'all', or list of 1D indices (optional)
            Default is None, which will read the sky fibre list from
            the `info` attribute; if absent or empty, it will trigger 'auto'.
            The 'auto' option will call `estimate_sky_fibres` to identify
            sky fibres based on a mean intensity threshold, automatically
            determined from the shape of the normalised spectra.
            Selecting `all` will use every spectrum in the `DataContainer`.
            Othrwise, this argument will be interpreted as a list of
            integer indices identifying sky spectra, assuming 1D RSS order.
        source_mask_nsigma : float, optional
            Sigma level for masking sources. Default is None.
        qc_plots: dict
            Dictionary to control QC plots.

        Returns
        -------
        np.ndarray
            Estimated background.
        np.ndarray
            Estimated background standard deviation.
        """
        if bckgr_params is None:
            bckgr_params = {}

        if hasattr(BackgroundEstimator, bckgr_estimator):
            estimator = getattr(BackgroundEstimator, bckgr_estimator)
        else:
            raise NameError(
                f"Input background estimator {bckgr_estimator} does not exist")
        
        # Determine sky fibres:
        
        self.sky_fibres = sky_fibres
        if sky_fibres is None:
            self.sky_fibres = self.dc.info.get('sky_fibres', None)
            if self.sky_fibres is None or len(self.sky_fibres) == 0:
                sky_fibres = 'auto'
        if sky_fibres == "auto":
            self.sky_fibres = self._estimate_sky_fibres()
        elif sky_fibres == "all":
            self.sky_fibres = np.arange(int(self.dc.n_spectra))
        #bckgr_params["axis"] = bckgr_params.get("axis", 0)
        #data = np.take(self.dc.rss_intensity, self.sky_fibres, bckgr_params["axis"])
        data = self.dc.rss_intensity[self.sky_fibres]
        
        if (len(qc_plots) > 0) and (self.dc.intensity.ndim == 2):
            show_plot = qc_plots.get('show', True)
            plot_filename = qc_plots.get('filename_base', None)
            fig_name = 'sky_fibres'
            fig, axes = new_figure(fig_name, figsize=(8, 6))
            fig.suptitle(f"{self.dc.info['name']} {fig_name}")
            total_flux = np.nanmean(self.dc.rss_intensity.value, axis=1)
            ax, patch_collection, cb = plot_fibres(
                fig, axes[0, 0], self.dc, data=total_flux,
                cblabel=f'mean intensity [{self.dc.rss_intensity.unit}]',
                cmap='Spectral_r',
                norm=Normalize(vmax=np.nanmean(total_flux)))
            handle, = ax.plot(self.dc.info['fib_ra'][self.sky_fibres],
                              self.dc.info['fib_dec'][self.sky_fibres],
                              'k+', label=f'{self.sky_fibres.size} sky fibres')
            ax.legend(handles=[handle])
            if plot_filename is not None:
                fig.savefig(f'{plot_filename}_{fig_name}.png')
            self.qc_plots[fig_name] = fig
            #if show_plot:
            #    plt.show(fig_name)
            #else:
            if not show_plot:
                plt.close(fig_name)

        # Estimate sky spectrum:
        
        if source_mask_nsigma is not None:
            vprint("Pre-estimating background using all data")
            bckgr, bckgr_sigma = estimator(data, **bckgr_params)
            vprint(
                f"Applying sigma-clipping mask (n-sigma={source_mask_nsigma})")
            dims_to_expand = (0)
            source_mask = (data > np.expand_dims(bckgr, dims_to_expand) +
                           source_mask_nsigma
                           * np.expand_dims(bckgr_sigma, dims_to_expand))
            data[source_mask] = np.nan

        vprint(f"Applying the {bckgr_estimator} estimator to {data.shape[0]} sky fibres")
        bckgr, bckgr_sigma = estimator(data, **bckgr_params)
        return bckgr, bckgr_sigma

    def _estimate_sky_fibres(self):
        """
        Identify sky fibres by imposing a mean intensity threshold, based on
        the shape of the normalised spectra.
        """
        
        total_flux = np.nanmean(self.dc.rss_intensity, axis=1).value
        sorted_by_flux = np.argsort(total_flux)
        n = 1 + np.arange(sorted_by_flux.size)
        half_sample = sorted_by_flux.size // 2

        flux_mean = np.nancumsum(total_flux[sorted_by_flux]) / n

        norm_fibre = self.dc.rss_intensity.value[sorted_by_flux, :] / total_flux[sorted_by_flux][:, np.newaxis]
        norm_sum1 = np.nancumsum(norm_fibre, axis=0)
        norm_sum2 = np.nancumsum(norm_fibre**2, axis=0)

        left_side_mean = norm_sum1[1:half_sample] / n[1:half_sample, np.newaxis]
        right_side_mean = norm_sum1[3::2] / n[1:half_sample, np.newaxis] - left_side_mean

        left_side_err = norm_sum2[1:half_sample] / n[1:half_sample, np.newaxis]
        right_side_err = norm_sum2[3::2] / n[1:half_sample, np.newaxis] - left_side_err

        left_side_err -= left_side_mean**2
        left_side_err[~(left_side_err > 0)] = np.nan
        left_side_err = np.sqrt(left_side_err / n[:half_sample-1, np.newaxis])  # n-1

        right_side_err -= right_side_mean**2
        right_side_err[~(right_side_err > 0)] = np.nan
        right_side_err = np.sqrt(right_side_err / n[:half_sample-1, np.newaxis])  # n-1

        weight = np.exp(np.nanmean(
            -(right_side_mean - left_side_mean)**2 * (1/right_side_err**2 + 1/left_side_err**2),
            axis=1))
        sky_flux = np.nansum(total_flux[sorted_by_flux][3::2] * weight) / np.nansum(weight)

        n_sky = min(
            np.searchsorted(flux_mean, sky_flux),
            2 * np.searchsorted(total_flux[sorted_by_flux], sky_flux))
        flux_threshold = total_flux[sorted_by_flux[n_sky-1]]
        sky_fibres = sorted_by_flux[:n_sky]
        vprint(f'{n_sky} sky fibres found below {flux_threshold:.5g} (sky flux = {sky_flux:.5g}) {self.dc.rss_intensity.unit}')
        
        return sky_fibres

    def plot_individual_wavelength(self, wavelength):
        """
        Identify sky fibres by imposing a mean intensity threshold, based on
        the shape of the normalised spectra.
        """
        idx = np.searchsorted(self.dc.wavelength, check_unit(wavelength, u.Angstrom)) - 1

        intensity = self.dc.rss_intensity[:, idx].value
        total_flux = np.nanmean(self.dc.rss_intensity, axis=1)
        intensity_norm = intensity / total_flux

        sorted_by_flux = np.argsort(total_flux)
        flux_mean = np.nancumsum(total_flux[sorted_by_flux]) / np.arange(1, total_flux.size+1)
        half_sample = total_flux.size // 2
        
        fig, axes = new_figure('single_wavelength_sky', nrows=2)

        ax = axes[0, 0]
        ax.set_ylabel(f'intensity ($\\lambda={self.dc.wavelength[idx]:.2f}\\ \\AA$)')
        vmin = np.nanmin(intensity[self.sky_fibres])
        vmax = np.nanmax(intensity[self.sky_fibres])
        h = .1 * (vmax - vmin)
        ax.set_ylim(vmin-h, vmax+2*h)

        ax.plot(total_flux, intensity, 'k.', alpha=.1)
        ax.plot(total_flux[self.sky_fibres], intensity[self.sky_fibres], 'b+', alpha=.5)
        ax.axhline(self.intensity[idx], c='k', ls='--')
        std = np.sqrt(self.variance[idx])
        ax.axhline(self.intensity[idx] + std, c='k', ls=':')
        ax.axhline(self.intensity[idx] - std, c='k', ls=':')


        ax = axes[1, 0]
        ax.set_ylabel(f'normalised intensity')
        vmin = np.nanmin(intensity_norm[self.sky_fibres])
        vmax = np.nanmax(intensity_norm[self.sky_fibres])
        h = .1 * (vmax - vmin)
        ax.set_ylim(vmin-h, vmax+2*h)

        ax.plot(total_flux, intensity_norm, 'k.', alpha=.1)
        ax.plot(total_flux[self.sky_fibres], intensity_norm[self.sky_fibres], 'b+', alpha=.5)
        sky_flux = np.nanmean(self.intensity.value)
        ax.axhline(self.intensity.value[idx]/sky_flux, c='k', ls='--')
        ax.axhline((self.intensity[idx] + std).value/sky_flux, c='k', ls=':')
        ax.axhline((self.intensity[idx] - std).value/sky_flux, c='k', ls=':')

        ax.set_xlabel('total flux (mean fibre intensity)')
        vmin, vmax = (flux_mean[0], flux_mean[-1])
        h = .1 * (vmax - vmin)
        ax.set_xlim(vmin-h, vmax+2*h)

        return fig

# =============================================================================
# Sky Substraction Correction
# =============================================================================


class SkySubsCorrection(CorrectionBase):
    """
    Correction for removing sky emission from a DataContainer.

    This class applies sky emission correction to a `DataContainer`
    using a provided sky model.
    It supports both standard and PCA-based sky subtraction methods
    and can generate visualizations of the correction process.

    Attributes
    ----------
    name : str
        Name of the correction.
    verbose : bool
        Flag to control verbosity of the correction process.
    skymodel : SkyModel
        The sky model used for sky emission subtraction.
    """

    name = "SkyCorrection"

    def __init__(self, skymodel, **correction_kwargs):
        """
        Initialize the SkySubsCorrection with a given sky model.

        Parameters
        ----------
        skymodel : SkyModel
            The sky model to be used for sky emission subtraction.
        """
        super().__init__(**correction_kwargs)
        self.skymodel = skymodel

    def plot_correction(self, data_cont, data_cont_corrected, **kwargs):
        """
        Plot the original and sky-corrected intensity of a DataContainer for comparison.

        Parameters
        ----------
        data_cont : :class:`pykoala.data_container.DataContainer`
            The original DC before sky correction.
        data_cont_corrected : :class:`pykoala.data_container.DataContainer`
            The DC after sky correction.
        kwargs : dict
            Additional keyword arguments for `imshow`.

        Returns
        -------
        fig : :class:`matplotlib.figure.Figure`
            The figure object containing the plots.
        """
        
        if data_cont.intensity.ndim == 2:
            original_image = data_cont.intensity
            corr_image = data_cont_corrected.intensity
        elif data_cont.intensity.ndim == 3:
            original_image = data_cont.get_white_image()
            data_unit = str(original_image.unit)
            original_image = original_image
            corr_image = data_cont_corrected.get_white_image()

        norm = plt.Normalize(np.nanpercentile(original_image.value, 1),
                             np.nanpercentile(original_image.value, 99))
        fig, axs = plt.subplots(nrows=1, ncols=2, figsize=(10, 5),
                                constrained_layout=True,
                                sharex=True, sharey=True)

        plot_image(fig, axs[0], data=original_image, norm=norm)
        plot_image(fig, axs[1], data=corr_image, norm=norm)
        axs[0].set_title("Original")
        axs[1].set_title("Sky emission subtracted")

        if kwargs.get("show", False):
            plt.show()
        else:
            plt.close(fig)
        return fig

    def apply(self, dc, pca=False, plot=False, **plot_kwargs):
        """
        Apply the sky emission correction to the datacube.

        Parameters
        ----------
        dc : :class:`pykoala.data_container.DataContainer`
            The DataContainer to be corrected.
        pca : bool, optional
            If True, use PCA-based sky subtraction. Default is False.
        verbose : bool, optional
            If True, print progress messages. Default is True.
        plot : bool, optional
            If True, generate and return plots of the correction. Default is False.
        plot_kwargs : dict
            Additional keyword arguments for the plot.

        Returns
        -------
        dc_out : :class:`pykoala.data_container.DataContainer`
            The corrected datacube.
        fig : matplotlib.figure.Figure or None
            The figure object containing the plots if `plot` is True, otherwise None.
        """
        # Set verbosity
        # Copy input datacube to store the changes
        dc_out = copy.deepcopy(dc)

        self.vprint("Applying sky subtraction")

        if pca:
            dc_out.intensity, dc_out.variance = self.skymodel.substract_pca(
                dc_out.intensity, dc_out.variance)
        else:
            dc_out.intensity, dc_out.variance = self.skymodel.substract(
                dc_out.intensity, dc_out.variance)

        self.record_correction(dc_out, status='applied')
        if plot:
            fig = self.plot_correction(dc, dc_out, **plot_kwargs)
        else:
            fig = None

        return dc_out, fig


# =============================================================================
# Telluric Correction
# =============================================================================


class TelluricCorrection(CorrectionBase):
    """
    Corrects for telluric absorption caused by atmospheric effects.

    This class implements methods to estimate and apply corrections for telluric absorption
    effects to data containers.

    Attributes
    ----------
    name : str
        The name of the correction method.
    telluric_correction : array
        The computed telluric correction.
    verbose : bool
        Controls verbosity of logging messages.

    Methods
    -------
    telluric_from_smoothed_spec(exclude_wlm=None, step=10,
                                weight_fit_median=0.5, wave_min=None,
                                wave_max=None, plot=True, verbose=False):
        Estimates the telluric correction from smoothed spectra.

    telluric_from_model(file='telluric_lines.txt', width=30,
                        extra_mask=None, pol_deg=5, plot=False):
        Estimates the telluric correction using a model of absorption lines.

    plot_correction(fig_size=12, wave_min=None, wave_max=None,
                    exclude_wlm=None, **kwargs):
        Plots the telluric correction.

    apply(rss, verbose=True, is_combined_cube=False, update=True):
        Applies the telluric correction to the input data.

    interpolate_model(wavelength, update=True):
        Interpolates the telluric correction model to match the input wavelength array.

    save(filename='telluric_correction.txt', **kwargs):
        Saves the telluric correction to a text file.
    """
    name = "TelluricCorretion"
    _telluric_correction = None
    _wavelength = None
    default_model_file = os.path.join(os.path.dirname(__file__), '..',
                                      'input_data', 'sky_lines',
                                      'telluric_lines.txt')
    
    @property
    def telluric_correction(self):
        return self._telluric_correction

    @telluric_correction.setter
    def telluric_correction(self, value):
        self._telluric_correction = value

    @property
    def wavelength(self):
        return self._wavelength

    @wavelength.setter
    def wavelength(self, value):
        self._wavelength = check_unit(value, u.angstrom)

    @property
    def airmass(self) -> float:
        """Airmass at which the model has been created."""
        return self._airmass
    
    @airmass.setter
    def airmass(self, value):
        self._airmass = value

    def __init__(self,
                 telluric_correction,
                 wavelength,
                 airmass,
                 telluric_correction_file="unknown",
                 **correction_kwargs):
        """
        Initializes the TelluricCorrection object.

        Parameters
        ----------
        telluric_correction : :class:`numpy.ndarray`
            Array containing the values of the telluric correction.
        wavelength : :class:`astropy.units.Quantity`
            Wavelength values associated to the ``telluric_correction``.
        airmass : float
            Airmass at which the model is estimated.
        telluric_correction_file : str, optional
            Path to the file that contains the model.
        **correction_kwargs:
            Additional arguments passed to :class:`CorrectionBase`.
        """
        super().__init__(**correction_kwargs)

        self.telluric_correction = telluric_correction
        self.wavelength = wavelength
        if self.telluric_correction.size != self.wavelength.size:
            raise ValueError("Size of telluric correction and input wavelength"
                             " vector are not compatible: "
                             f"{self.telluric_correction.size}, {self.wavelength.size}")
        self.airmass = airmass
        self.telluric_correction_file = telluric_correction_file

    @classmethod
    def from_text_file(cls, path):
        """Initialise a TelluricCorrection from a text file.
        
        Parameters
        ----------
        path : str
            Path to the text file containing the telluric correction. The first
            and second columns must contain the wavelength and telluric correction,
            respectively.

        Returns
        -------
        telluric_correction : :class:`TelluricCorrection`
            The telluric correction.
        """
        vprint("Initialising telluric correction from text file")
        wavelength, telluric_correction = np.loadtxt(path, unpack=True,
                                                     usecols=(0, 1))
        # Read the value of the airmass at which the model was computed
        with open(path, "r") as f:
            line = f.readline()
            if "airmass" in line:
                airmass = float(line.split("=")[1])
            else:
                raise ValueError("Telluric correction file must include"
                                 " airmass=value in the first line")
        return cls(telluric_correction=telluric_correction,
                   wavelength=wavelength,
                   telluric_correction_file=path, airmass=airmass)

    @classmethod
    def from_smoothed_spectra_container(cls, spectra_container,
                                        exclude_wlm=None, 
                                        light_percentile=0.95,
                                        median_window=10 << u.angstrom,
                                        wave_min=None, wave_max=None,
                                        plot=True):
        """
        Estimate the telluric correction function using the smoothed spectra of the input star.

        Parameters
        ----------
        spectra_container: `pykoala.data_container.SpectraContainer`
            SpectraContainer used to estimate the telluric absorption.
        exclude_wlm : list of lists, optional
            List of wavelength ranges to exclude from correction (default is None).
        light_percentile: float, optional
            Percentile of light used to estimate the absorption. Default is 0.95.
        median_window : int, optional
            Step size for smoothing (default is 10).
        wave_min : float, optional
            Minimum wavelength to consider (default is None).
        wave_max : float, optional
            Maximum wavelength to consider (default is None).
        plot : bool, optional
            Whether to plot the correction (default is True).

        Returns
        -------
        telluric_correction : `TelluricCorrection`
            The computed telluric correction.
        fig : :class:`matplotlib.pyplot.Figure` or ``None``
            If ``plot=True``, it corresponds to a quality control plot.
        """
        vprint("Initialising telluric correction from input STD star")
        spectra = np.nanpercentile(spectra_container.rss_intensity,
                                   light_percentile, axis=0)
        telluric_correction = np.ones(spectra_container.wavelength.size)
        if wave_min is None:
            wave_min = spectra_container.wavelength[0]
        if wave_max is None:
            wave_max = spectra_container.wavelength[-1]
        if exclude_wlm is None:
            exclude_wlm = np.array([[6450 , 6700], [6850, 7050], [7130, 7380]]
                                   ) << u.angstrom
        # Mask containing the spectral points to include in the telluric correction
        correct_mask = (spectra_container.wavelength >= wave_min) & (
                        spectra_container.wavelength <= wave_max)
        # Mask user-provided spectral regions
        spec_windows_mask = np.ones_like(correct_mask, dtype=bool)
        for window in exclude_wlm:
            spec_windows_mask[(spectra_container.wavelength >= window[0]) &
                              (spectra_container.wavelength <= window[1])
                              ] = False
        # Master mask used to compute the Telluric correction
        telluric_mask = correct_mask & spec_windows_mask
        if not median_window % 2:
            median_window += 1
        smooth_med_star = ContinuumEstimator.medfilt_continuum(spectra,
                                                               median_window)
        telluric_correction[telluric_mask] = (smooth_med_star[telluric_mask]
                                              / spectra[telluric_mask])

        if plot:
            fig = TelluricCorrection.plot_correction(
                spectra_container, telluric_correction,
                wave_min=wave_min, wave_max=wave_max, exclude_wlm=exclude_wlm)
        else:
            fig = None
        return cls(telluric_correction=telluric_correction,
                   wavelength=spectra_container.wavelength,
                   airmass=spectra_container.info["airmass"]), fig

    @classmethod
    def from_model(cls, spectra_container, model_file=None,
                   width=30, extra_mask=None, plot=False):
        """
        Estimate the telluric correction function using a model of telluric absorption lines.

        Parameters
        ----------
        file : str, optional
            Path to the file containing telluric lines (default is 'telluric_lines.txt').
        width : int, optional
            Half-window size to account for instrumental dispersion (default is 30).
        extra_mask : array, optional
            Mask of additional spectral regions to exclude (default is None).
        pol_deg : int, optional
            Polynomial degree for fitting (default is 5).
        plot : bool, optional
            Whether to plot the correction (default is False).

        Returns
        -------
        telluric_correction : `TelluricCorrection`
            The computed telluric correction.
        fig : :class:`matplotlib.pyplot.Figure` or ``None``
            If ``plot=True``, it corresponds to a quality control plot.
        """
        if model_file is None:
            model_file = TelluricCorrection.default_model_file

        width = check_unit(width, u.angstrom)
        w_l_1, w_l_2 = np.loadtxt(model_file, unpack=True, usecols=(0, 1))
        w_l_1 = w_l_1 << u.angstrom
        w_l_2 = w_l_2 << u.angstrom
        # Mask telluric regions
        mask = np.ones(spectra_container.wavelength.size, dtype=bool)
        telluric_correction = np.ones(spectra_container.wavelength.size,
                                      dtype=float)
        for b, r in zip(w_l_1, w_l_2):
            mask[(spectra_container.wavelength >= b - width) & (
                spectra_container.wavelength <= r + width)] = False
        if extra_mask is not None:
            mask = mask & extra_mask
        std = np.nanstd(spectra_container.rss_intensity, axis=0)
        stellar = np.interp(spectra_container.wavelength,
                            spectra_container.wavelength[mask], std[mask])
        telluric_correction[~mask] = stellar[~mask] / std[~mask]
        telluric_correction = np.clip(telluric_correction, a_min=1, a_max=None)
        if plot:
            fig = TelluricCorrection.plot_correction(
                spectra_container, telluric_correction,
                exclude_wlm=np.vstack((w_l_1 - width, w_l_2 + width)).T,
                wave_min=spectra_container.wavelength[0],
                wave_max=spectra_container.wavelength[-1])
        else:
            fig = None
        return cls(telluric_correction=telluric_correction,
                   wavelength=spectra_container.wavelength,
                   airmass=spectra_container.info["airmass"]), fig

    @classmethod
    def flag_data_container(cls, data_container, telluric_correction=None,
                            wavelength=None, path_to_model=None,
                            min_line_width=5 << u.AA,
                            flag_name="telluric"):
        """Flag the pixels of a data container affected by telluric absorption.
        
        Parameters
        ----------
        data_container : :class:`DataContainer`
            Input DataContainer
        telluric_correction : np.array, optional
            Input telluric correction values. Regions affected by telluric
            absortion should present values larget than 1, and 1 otherwise.
        wavelength : np.ndarray, optional
            Wavelength associated to the telluric correction.
        path_to_model : str
            Path to a text file containing the left and right edges of each
            telluric line.
        """

        telluric_flag = np.zeros(data_container.rss_intensity.shape,
                                 dtype=bool)
        if telluric_correction is not None:
            if wavelength is None:
                raise ValueError(
                    "Must provide a wavelength associated to the correction")
            else:
                interp_tell_corr = np.interp(
                    data_container.wavelength, wavelength, telluric_correction)
                telluric_flag[:, interp_tell_corr >= 1.001] = True
        else:
            if path_to_model is None:
                path_to_model = cls.default_model_file
        
            telluric_wl1, telluric_wl2 = np.loadtxt(
                path_to_model, unpack=True, usecols=(0, 1))
            telluric_wl1 = telluric_wl1 << u.AA
            telluric_wl2 = telluric_wl2 << u.AA
            for b, r in zip(telluric_wl1, telluric_wl2):
                # If the line is narrower than the minimum required
                if r - b < min_line_width:
                    d = min_line_width - (r - b)
                    r += d / 2
                    b -= d / 2
                b_idx = data_container.wcs.spectral.world_to_array_index(
                    b)
                r_idx = data_container.wcs.spectral.world_to_array_index(
                    r)
                if r_idx == b_idx:
                    r_idx += 1
                telluric_flag[:, slice(b_idx, r_idx)] = True
        telluric_flag = data_container.rss_to_original(telluric_flag)
        data_container.mask.flag_pixels(telluric_flag, flag_name,
                                        desc="telluric absoption contaminated")

    @staticmethod
    def plot_correction(spectra_container, telluric_correction,
                        wave_min=None, wave_max=None, exclude_wlm=None,
                        **kwargs):
        """
        Plot a telluric correction.

        Parameters
        ----------
        wave_min : float, optional
            Minimum wavelength to display (default is None).
        wave_max : float, optional
            Maximum wavelength to display (default is None).
        exclude_wlm : array, optional
            List of wavelength ranges to exclude from plot (default is None).
        **kwargs
            Additional keyword arguments for plot customization.

        Returns
        -------
        fig : :class:`matplotlib.pyplot.Figure`
            The matplotlib figure object.
        """
        fig, ax = plt.subplots()
        sorted_idx = spectra_container.get_spectra_sorted()
        ax.set_title("Telluric correction using fibres {} (cyan) and {} (fuchsia)"
                        .format(sorted_idx[-2], sorted_idx[-1]))

        ax.plot(spectra_container.wavelength,
                spectra_container.rss_intensity[sorted_idx[-2]], color="c",
                label='Original', lw=3, alpha=0.8)
        ax.plot(spectra_container.wavelength,
                spectra_container.rss_intensity[sorted_idx[-2]]
                * telluric_correction, color="deepskyblue",
                label='Corrected', lw=1, alpha=0.8)

        ax.plot(spectra_container.wavelength,
                spectra_container.rss_intensity[sorted_idx[-1]], color="fuchsia",
                label='Original', lw=3, alpha=0.8)
        ax.plot(spectra_container.wavelength,
                spectra_container.rss_intensity[sorted_idx[-1]]
                * telluric_correction, color="purple",
                label='Corrected', lw=1, alpha=0.8)
        ax.set_ylim(np.nanpercentile(
            spectra_container.rss_intensity[sorted_idx[-1]], [1, 99]).value)
        ax.set_ylabel(f"Flux ({spectra_container.intensity.unit})")
        ax.legend(ncol=2)
        ax.axvline(x=wave_min.to_value(spectra_container.wavelength.unit),
                   color='lime', linestyle='--')
        ax.axvline(x=wave_max.to_value(spectra_container.wavelength.unit),
                   color='lime', linestyle='--')
        ax.set_xlabel(r"Wavelength [$\mathrm{\AA}$]")
        if exclude_wlm is not None:
            for i in range(len(exclude_wlm)):
                ax.axvspan(
                    exclude_wlm[i][0].to_value(spectra_container.wavelength.unit),
                    exclude_wlm[i][1].to_value(spectra_container.wavelength.unit),
                    color='lightgreen', alpha=0.1)
        ax.minorticks_on()
        ax.set_xlim(spectra_container.wavelength[[0, -1]].value)
        if kwargs.get('plot', False):
            plt.show()
        else:
            plt.close()
        return fig

    def apply(self, spectra_container, update=True):
        """
        Apply the telluric correction to the input :class:`SpectraContainer`.

        Parameters
        ----------
        spectra_container : :class:`SpectraContainer`
            The input SpectraContainer to correct.
        update : bool, optional
            Whether to update the correction (default is True).

        Returns
        -------
        rss_out : :class:`SpectraContainer`
            The corrected copy of the input SpectraContainer.
        """

        # Check wavelength
        if not spectra_container.wavelength.size == self.wavelength.size or not np.allclose(
            spectra_container.wavelength, self.wavelength, equal_nan=True):
            self.vprint("Interpolating correction to input wavelength")
            tell_correction = self.interpolate_model(
                spectra_container.wavelength, update=update)
        else:
            tell_correction = self.telluric_correction

        if spectra_container.is_corrected(self.name):
            self.vprint("Data already calibrated")
            return spectra_container

        # Copy input RSS for storage the changes implemented in the task
        tell_correction = tell_correction**(
            spectra_container.info["airmass"] / self.airmass)
        spectra_container_out = spectra_container.copy()
        self.vprint("Applying telluric correction")
        spectra_container_out.rss_intensity = (spectra_container_out.rss_intensity
                                               * tell_correction)
        spectra_container_out.rss_variance = (spectra_container_out.rss_variance
                                              * tell_correction**2)
        self.record_correction(spectra_container_out, status='applied')
        return spectra_container_out

    def interpolate_model(self, wavelength, update=True):
        """
        Interpolate the telluric correction model to match the input wavelength array.

        Parameters
        ----------
        wavelength : array
            The wavelength array to interpolate.
        update : bool, optional
            Whether to update the correction (default is True).

        Returns
        -------
        telluric_correction : array
            The interpolated telluric correction.
        """
        telluric_correction = np.interp(
            wavelength, self.wavelength, self.telluric_correction,
            left=1, right=1)

        if update:
            self.telluric_correction = telluric_correction
            self.wavelength = wavelength
        return telluric_correction

    def save(self, filename='telluric_correction.txt', **kwargs):
        """
        Save the telluric correction to a text file.

        Parameters
        ----------
        filename : str, optional
            The name of the output file.
        **kwargs : dict
            Extra arguments to be passed to :func:`numpy.savetxt`.
        """
        self.vprint(f"Saving telluric correction into file {filename}")
        if "header" in kwargs:
            kwargs["header"] = f"airmass={self.airmass}\n" + kwargs["header"]
        else:
            kwargs["header"] = f"airmass={self.airmass}\n"

        np.savetxt(filename, np.array(
            [self.wavelength, self.telluric_correction]).T, **kwargs)


def combine_telluric_corrections(list_of_telcorr, ref_wavelength):
    """Combine a list of input telluric corrections.
    
    Parameters
    ----------
    list_of_telcorr : list
        List of :class:`TelluricCorrection` instances to combine.
    ref_wavelength : 1D np.ndarray
        Reference array grid to interpolate each correction before the combination.

    Returns
    -------
    combine_telcorr : :class:`TelluricCorrection`
        TelluricCorretion resulting from combining the input list.
    """
    vprint("Combining input telluric corrections")
    telluric_corrections = np.zeros(
        (len(list_of_telcorr), ref_wavelength.size))
    airmass = np.zeros(len(list_of_telcorr), dtype=float)
    for i, telcorr in enumerate(list_of_telcorr):
        telluric_corrections[i] = telcorr.interpolate_model(ref_wavelength)
        airmass[i] = telcorr.airmass

    telluric_correction = np.nanmedian(telluric_corrections, axis=0)
    return TelluricCorrection(telluric_correction=telluric_correction,
                              wavelength=ref_wavelength,
                              airmass=np.nanmedian(airmass), verbose=False)


# =============================================================================
# Self-calibration based on strong sky lines
# =============================================================================

# TODO: rename SkyWaveletFilter?

class WaveletFilter(object):
    '''
    Estimate overall fibre throughput and wavelength calibration based on sky emission lines (from wavelet transform).

    Description
    -----------

    Given a Row-Stacked Spectra (RSS) object:
    1. Estimate the FWHM of emission lines from the autocorrelation of the median (~ sky) spectrum.
    2. Apply a (mexican top hat) wavelet filter to detect features on that scale (i.e. filter out the continuum).
    3. Find regions actually dominated by sky lines (i.e. exclude lines from the target).
    4. Estimate fibre throughput from norm (standard deviation).
    5. Estimate wavelength offset of each fibre (in pixels) from cross-correlation with the sky.

    Attributes
    ----------
    - scale: (int)
        Scale of the wavelet filter, in pixels.
    - filtered: numpy.ndarray(float)
        Throughput-corrected wavelet coefficient (filtered rss.intensity).
    - sky_lo: numpy.ndarray(float)
        Lower percentile (16) of the estimated filtered sky.
    - sky: numpy.ndarray(float)
        Estimated (median) filtered sky.
    - sky_hi: numpy.ndarray(float)
        Upper percentile (84) of the estimated filtered sky.
    - sky_weight: numpy.ndarray(float)
        Estimated dominance of the sky (between 0 and 1).
    - fibre_throughput: numpy.ndarray(float)
        Overall throughput of each fibre.
    - wavelength: numpy.ndarray(float)
        Wavelength of the wavelet coefficients (only used for plotting).
    - fibre_offset: numpy.ndarray(float)
        Relative wavelenght calibration. Offset, in pixels, with respect to the sky (from weighted cross-correlation).
    '''

    def __init__(self, rss: RSS):

        # 1. Estimate the FWHM of emission lines from the autocorrelation of the median (~ sky) spectrum.

        x = np.nanmedian(rss.intensity, axis=0)  # median ~ sky spectrum
        x -= np.nanmean(x)
        x = scipy.signal.correlate(x, x, mode='same')
        h = (np.count_nonzero(x > 0.5*np.nanmax(x)) + 1) // 2
        # h = 0
        self.scale = 2*h + 1
        vprint(f'> Wavelet filter scale: {self.scale} pixels')

        # 2. Apply a (mexican top hat) wavelet filter to detect features on that scale (i.e. filter out the continuum).

        x = np.nancumsum(rss.intensity, axis=1)
        self.filtered = (x[:, 2*self.scale:-self.scale] -
                         x[:, self.scale:-2*self.scale]) / self.scale
        self.filtered -= (x[:, 3*self.scale:] -
                          x[:, :-3*self.scale]) / (3*self.scale)

        # 3. Find regions actually dominated by sky lines (i.e. exclude lines from the target).

        self.sky_lo, self.sky, self.sky_hi = np.nanpercentile(
            self.filtered, [16, 50, 84], axis=0)
        self.sky_weight = 1 - \
            np.exp(-.5 * (self.sky / np.fmax((self.sky - self.sky_lo),
                   (self.sky_hi - self.sky)))**2)
        self.sky *= self.sky_weight
        self.sky_lo *= self.sky_weight
        self.sky_hi *= self.sky_weight

        # 4. Estimate fibre throughput from norm (standard deviation).

        # should be irrelevant
        self.filtered -= np.nanmean(self.filtered, axis=1)[:, np.newaxis]
        self.filtered *= self.sky_weight
        # self.fibre_throughput = np.nanstd(self.filtered, axis=1)
        # self.fibre_throughput = np.nanmedian(self.filtered, axis=1)
        # self.filtered /= self.fibre_throughput[:, np.newaxis]
        self.filtered[~ np.isfinite(self.filtered)] = 0

        # self.fibre_throughput = np.nanmedian(self.filtered / self.sky, axis=1)
        x = np.exp(-.5 * ((self.filtered - self.sky) /
                   (self.sky_hi - self.sky_lo))**2)
        x *= self.sky_weight[np.newaxis, :]
        # self.fibre_throughput = np.nanmean(x * self.filtered / self.sky[np.newaxis, :], axis=1) / np.nanmean(x)
        x = np.where(x > 0.5, self.filtered / self.sky[np.newaxis, :], np.nan)
        self.fibre_throughput = np.nanmedian(x, axis=1)
        renorm = np.nanmedian(self.fibre_throughput)
        self.fibre_throughput /= renorm
        self.filtered /= self.fibre_throughput[:, np.newaxis]
        self.sky *= renorm
        self.sky_lo *= renorm
        self.sky_hi *= renorm

        # 5. Estimate wavelength offset of each fibre (in pixels) from cross-correlation with the sky.

        # only for plotting:
        x = u.Quantity(rss.wavelength[self.scale + h + 1: -self.scale - h])
        if x.unit.is_equivalent(u.AA):
            self.wavelength = x.to_value(u.AA)
        # assume it's already in Angstrom
        elif x.unit.is_equivalent(u.dimensionless_unscaled):
            self.wavelength = x.value
        else:
            raise TypeError(f'  ERROR: wrong wavelength units ({x.unit})')

        # mid = rss.intensity.shape[1] // 2
        mid = self.wavelength.size // 2
        s = self.scale
        x = np.nanmedian(self.filtered, axis=0)
        x[~ np.isfinite(x)] = 0
        x = scipy.signal.fftconvolve(
            self.filtered, x[np.newaxis, ::-1], mode='same', axes=1)[:, mid-s:mid+s+1]
        idx = np.arange(x.shape[1])
        weight = np.where(x > 0, x, 0)
        self.fibre_offset = np.nansum(
            (idx - s)[np.newaxis, :] * weight, axis=1) / np.nansum(weight, axis=1)

    def qc_plots(self, show=False, save_as=None):
        '''
        Top: Filtered intensity for every fibre.
        Centre: Sky-subtracted.
        Bottom: Filtered sky.
        '''
        fig, axes = new_figure('wavelet_filter', nrows=3, ncols=2, sharey=False, gridspec_kw={
                               'width_ratios': [1, .02], 'hspace': 0., 'wspace': .1})
        im, cb = plot_image(fig, axes[0, 0], 'wavelet coefficient', self.filtered,
                            x=self.wavelength, ylabel='spec_id', cbax=axes[0, 1])
        im, cb = plot_image(fig, axes[1, 0], 'sky-subtracted', self.filtered -
                            self.sky, x=self.wavelength, ylabel='spec_id', cbax=axes[1, 1])
        axes[1, 0].sharey(axes[0, 0])
        ax = axes[-1, 0]
        axes[-1, 1].axis('off')
        ax.plot(self.wavelength, self.sky, alpha=.5, c='tomato',
                label='sky (median coefficient)')
        ax.fill_between(self.wavelength, self.sky_lo, self.sky_hi, color='r',
                        alpha=0.1, label='uncertainty (16-84%)')
        ax.plot(self.wavelength, self.sky/self.sky_weight,
                alpha=.5, label='unweighted sky')
        ax.set_xlabel('wavelength [pix]')
        ax.set_ylabel(f'filtered (scale = {self.scale} pix)')
        ax.legend()
        if show:  # TODO: deal with ipympl
            fig.show()
        if save_as is not None:
            fig.savefig(save_as)

    def get_throughput_object(self):
        return Throughput(throughput_data=np.repeat(self.fibre_throughput[:, np.newaxis], self.wavelength.size + 3*self.scale, axis=1),
                          throughput_error=np.full([self.fibre_throughput.size, self.wavelength.size + 3*self.scale], np.nan)
                         )

    def get_wavelength_offset(self):
        return WavelengthOffset(offset_data=np.repeat(self.fibre_offset[:, np.newaxis], self.wavelength.size + 3*self.scale, axis=1) << u.pix)


# =============================================================================
# Mr Krtxo \(ﾟ▽ﾟ)/
#                                                       ... Paranoy@ Rulz! ;^D



"""
This script contains the wrapper functions to build a PyKoala RSS object from Hector (2dfdr-reduced) data.
"""

# =============================================================================
# Basics packages
# =============================================================================
import numpy as np
import os
import copy
# =============================================================================
# Astropy and associated packages
# =============================================================================
from astropy.io import fits
from astropy.wcs import WCS
# =============================================================================
# KOALA packages
# =============================================================================
from pykoala import vprint  # Template to create the info variable 
from pykoala.instruments.koala_ifu import read_rss


def airmass_from_header(header):
    """
    Compute the airmass extracting the parameters from KOALAS's header'
    """
    # Get ZD, airmass
    ZDSTART = header['ZDSTART']
    ZDEND = header['ZDEND']
    ZD = (ZDSTART + ZDEND) / 2
    airmass = 1 / np.cos(np.radians(ZD))
    return airmass


def hector_header(header):
    """
    Copy 2dfdr headers values from extensions 0 and 2 needed for the initial
    header for PyKoala. (based in the header constructed in  save_rss_fits in
    koala.io)
    """
    # To fit actual PyKoala header format
    header.rename_keyword('CENRA', 'RACEN')
    header.rename_keyword('CENDEC', 'DECCEN')
    return header

def set_fibre_table(fibre_table, bundle=''):
    """
    Generates the spaxels tables needed for PyKoala from the 2dfdr spaxels table.
    """
    # Filtering only selected (in use) fibres
    bundle_fibres = (fibre_table['PROBENAME'] == bundle) & (fibre_table['SELECTED'] == 1)
    bad_fibres = np.where(fibre_table['PROBENAME'] != bundle)[0]
    # Update the table
    fibre_table = fibre_table[bundle_fibres]

    # Defining new arrays
    arr1 = np.arange(len(fibre_table)) + 1  # +  for starting in 1
    arr2 = np.ones(len(fibre_table))
    arr3 = np.ones(len(fibre_table))
    arr4 = np.ones(len(fibre_table)) * 2048
    arr5 = np.zeros(len(fibre_table))
    arr6 = fibre_table['XPOS']
    arr7 = fibre_table['YPOS']
    arr8 = fibre_table['SPEC_ID']

    # Defining new columns
    col1 = fits.Column(name='Fibre', format='I', array=arr1)
    col2 = fits.Column(name='Status', format='I', array=arr2)
    col3 = fits.Column(name='Ones', format='I', array=arr3)
    col4 = fits.Column(name='Wavelengths', format='I', array=arr4)
    col5 = fits.Column(name='Zeros', format='I', array=arr5)
    col6 = fits.Column(name='Delta_RA', format='D', array=arr6)
    col7 = fits.Column(name='Delta_Dec', format='D', array=arr7)
    col8 = fits.Column(name='Fibre_OLD', format='I', array=arr8)

    # PyKoala Spaxels table
    fibre_table = fits.BinTableHDU.from_columns([col1, col2, col3, col4, col5, col6, col7, col8])

    return fibre_table, bad_fibres


def hector_rss(path_to_file, bundle=''):
    """
    A wrapper function that converts a file (not an RSS object) to a koala RSS object
    The paramaters used to build the RSS object e.g. bad spaxels, header etc all come from the original (non PyKoala) .fits file
    """
    header = fits.getheader(path_to_file, 0) + fits.getheader(path_to_file, 2)
    header = hector_header(header)
    # WCS
    wcs = WCS(header)
    # Constructing Pykoala Spaxels table from 2dfdr spaxels table (data[2])
    fibre_table = fits.getdata(path_to_file, 2)
    koala_fibre_table, bad_fibres = set_fibre_table(fibre_table, bundle=bundle)

    # List of bad spaxels from 2dfdr spaxels table
    # -1 to start in 0 rather than in 1
    # Create the dictionary containing relevant information
    info = {}
    info['name'] = header['OBJECT']
    info['exptime'] = header['EXPOSED']
    info['obj_ra'] = None
    info['obj_dec'] = None
    info['cen_ra'] = np.rad2deg(header['RACEN'])
    info['cen_dec'] = np.rad2deg(header['DECCEN'])
    info['pos_angle'] = header.get('TEL_PA', 0)
    info['fib_ra'] = koala_fibre_table.data['Delta_RA']
    info['fib_dec'] = koala_fibre_table.data['Delta_DEC']
    info['airmass'] = airmass_from_header(header)
    # Read RSS file into a PyKoala RSS object
    rss = read_rss(path_to_file, wcs=wcs,
                   bad_fibres_list=bad_fibres,
                   header=header,
                   fibre_table=koala_fibre_table,
                   info=info,
                   )
    return rss

# Mr Krtxo \(ﾟ▽ﾟ)/



"""
Module for estimating and applying wavelength offset corrections related to
inaccuracies in the original wavelength calibration.
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.gridspec import GridSpec

from astropy.io import fits
from astropy import units as u
from scipy.ndimage import (median_filter, gaussian_filter, percentile_filter,
                           label, labeled_comprehension)
from scipy.interpolate import interp1d
from scipy.signal import correlate, correlation_lags, convolve

from pykoala import vprint
from pykoala.corrections.correction import CorrectionBase
from pykoala.data_container import RSS
from pykoala.ancillary import flux_conserving_interpolation, vac_to_air, check_unit
from pykoala import ancillary


class WavelengthOffset(object):
    """Wavelength offset class.

    This class stores a 2D wavelength offset.

    Attributes
    ----------
    offset_data : :class:`astropy.units.Quantity`
        Wavelength offset expressed in pixels or wavelengths.
    offset_error : :class:`astropy.units.Quantity`
        Standard deviation of ``offset_data``.
    path: str
        Filename path.

    """
    offset_data = None
    offset_error = None

    def __init__(self, path=None, offset_data=None, offset_error=None):
        self.path = path
        # The input units can be either pixel or wavelength
        self.offset_data = check_unit(offset_data)
        self.offset_error = check_unit(offset_error)

    def to_fits(self, output_path=None):
        """Save the offset in a FITS file.
        
        Parameters
        ----------
        output_path: str, optional, default=None
            FITS file name path. If None, and ``self.path`` exists,
            the original file is overwritten.

        Notes
        -----
        The output fits file contains an empty PrimaryHDU, and two ImageHDU
        ("OFFSET", "OFFSET_ERR") containing the offset data and associated error.
        """
        if output_path is None:
            if self.path is None:
                raise NameError("Provide output path")
            else:
                output_path = self.path
        primary = fits.PrimaryHDU()
        header = fits.Header()
        header["bunit"] = self.offset_data.unit.to_string()
        data = fits.ImageHDU(data=self.offset_data.value, name='OFFSET', header=header)
        error = fits.ImageHDU(data=self.offset_error.value, name='OFFSET_ERR',
                              header=header)
        hdul = fits.HDUList([primary, data, error])
        hdul.writeto(output_path, overwrite=True)
        hdul.close(verbose=True)
        vprint(f"Wavelength offset saved at {output_path}")

    @classmethod
    def from_fits(cls, path):
        """Load the offset data from a fits file.

        Loads offset values (extension 1) and
        associated errors (extension 2) from a fits file.

        Parameters
        ----------
        path : str
            Path to the FITS file containing the offset data.

        Returns
        -------
        wavelength_offset : :class:`WavelengthOffset`
            A :class:`WavelengthOffset` initialised with the input data.
        """
        if not os.path.isfile(path):
            raise NameError(f"offset file {path} does not exist.")
        vprint(f"Loading wavelength offset from {path}")
        with fits.open(path) as hdul:
            offset_data = hdul[1].data << u.Unit(hdul[1].header.get("BUNIT", 1))
            offset_error = hdul[2].data << u.Unit(hdul[1].header.get("BUNIT", 1))
        return cls(offset_data=offset_data, offset_error=offset_error,
                   path=path)


class WavelengthCorrection(CorrectionBase):
    """Wavelength correction class.

    This class accounts for the relative wavelength offset between fibres.

    Attributes
    ----------
    name : str
        Correction name, to be recorded in the log.
    offset : :class:`WavelengthOffset`
        2D wavelength offset (n_fibres x n_wavelengths)
    verbose: bool
        False by default.
    """
    name = "WavelengthCorrection"
    offset = None
    verbose = False

    def __init__(self, offset_path: str=None, offset: WavelengthOffset=None,
                 **correction_kwargs):
        super().__init__(**correction_kwargs)

        self.path = offset_path
        self.offset = offset

    @classmethod
    def from_fits(cls, path):
        """Initialise a WavelegnthOffset correction from an input FITS file.
        
        Parameters
        ----------
        path : str
            Path to the FITS file containing the offset data.

        Returns
        -------
        wave_correction : :class:`WavelengthCorrection`
            A :class:`WavelengthCorrection` initialised with the input data.
        """
        return cls(offset=WavelengthOffset.from_fits(path=path),
                   offset_path=path)

    def apply(self, rss):
        """Apply a 2D wavelength offset model to a RSS.

        Parameters
        ----------
        rss : :class:`pykoala.rss.RSS`
            Original Row-Stacked-Spectra object to be corrected.

        Returns
        -------
        rss_corrected : :class:`pykoala.rss.RSS`
            Corrected copy of the input RSS.
        """

        assert isinstance(rss, RSS)

        rss_out = rss.copy()
        self.vprint("Applying correction to input RSS")
        if self.offset.offset_data.unit is u.pixel:
            x = np.arange(rss.wavelength.size) << u.pixel
        elif self.offset.offset_data.unit.is_equivalent(u.AA):
            x = rss.wavelength.to(self.offset.offset_data.unit)

        for i in range(rss.intensity.shape[0]):
                rss_out.intensity[i] = flux_conserving_interpolation(
                    x, x - self.offset.offset_data[i], rss.intensity[i])
                
        self.record_correction(rss_out, status='applied')
        return rss_out


class TelluricWavelengthCorrection(WavelengthCorrection):
    """WavelengthCorrection based on the cross-correlation of telluric lines.
    
    This class constructs a WavelengthOffset using a cross-correlation of
    telluric absorption lines.

    """
    name = "TelluricWavelengthCorrection"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    @classmethod
    def from_rss(cls, rss,
                 median_smooth=None, pol_fit_deg=None, oversampling=5,
                 wave_range=None, plot=False):
        """Estimate the wavelength offset from an input data container.
        
        Parameters
        ----------
        rss : :class:`RSS`
        median_smooth : int, optional
            Median filter size.
        pol_fit_deg : int, optional
            Polynomial degree to fit the resulting offset.
        oversampling : float, optional
            Oversampling factor to increase the accuracy of the cross-correlation.
        wave_range : list or tupla
            Wavelength range to fit the offset.
        plot : bool, optional
            If True, returns quality control plots.
        Returns
        -------
        WavelengthCorrection : :class:`WavelengthCorrection`
        figs : :class:`plt.Figure`
        """
        assert isinstance(rss, RSS), "Input data must be an instance of RSS"

        # mask = rss.mask.get_flag_map(telluric_flat)[0]
        # if not mask.any():
        #     raise ValueError("No telluric lines present in the data")
        # if wave_range is not None:
        #     mask &= rss.wavelength >= check_unit(
        #         wave_range[0], rss.wavelength.unit)
        #     mask &= rss.wavelength <= check_unit(
        #         wave_range[1], rss.wavelength.unit)

        intensity = rss.intensity.value.copy()
        intensity /= np.nanmedian(intensity, axis=1)[:, np.newaxis]
        # Emphasize the regions dominated by telluric absorption
        new_wavelength = np.interp(
            np.arange(0, rss.wavelength.size, 1 / oversampling),
            np.arange(rss.wavelength.size),
            rss.wavelength)
        interpolator = interp1d(rss.wavelength, intensity, axis=1)
        intensity = interpolator(new_wavelength)

        median_intensity = np.nanmedian(intensity, axis=0)
        fibre_offset = np.zeros(intensity.shape[0])
        for ith, fibre in enumerate(intensity):
            fibre_mask = np.isfinite(fibre)
            if not fibre_mask.any():
                continue

            corr = correlate(fibre[fibre_mask],
                             median_intensity[fibre_mask],
                             mode="full", method="fft")
            lags = correlation_lags(fibre[fibre_mask].size,
                                    median_intensity[fibre_mask].size)
            max_corr = np.argmax(corr)
            parabolic_max_lag = ancillary.parabolic_maximum(
                lags[[max_corr - 1, max_corr, max_corr + 1]],
                corr[[max_corr - 1, max_corr, max_corr + 1]])
            fibre_offset[ith] = parabolic_max_lag / oversampling

        if median_smooth is not None:
                fibre_offset = median_filter(fibre_offset, median_smooth)

        if pol_fit_deg is not None:
                pol = np.polyfit(np.arange(fibre_offset.size), fibre_offset,
                        deg=pol_fit_deg)
                fibre_offset = np.poly1d(pol)(np.arange(fibre_offset.size))            
        if plot:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(fibre_offset)


            ax.plot(fibre_offset)
            ax.set_ylim(fibre_offset.min(), fibre_offset.max())
            ax.set_xlabel("Fibre number")
            ax.set_ylabel(f"Average wavelength offset (pixel)")
            fibre_map_fig = rss.plot_fibres(data=fibre_offset)
            fig = (fig, fibre_map_fig)
        else:
            fig = None

        fibre_offset = fibre_offset << u.pixel
        offset = WavelengthOffset(
            offset_data=fibre_offset,
            offset_error=np.full_like(fibre_offset, fill_value=np.nan))
        return cls(offset=offset), fig


class SolarCrossCorrOffset(WavelengthCorrection):
    """WavelengthCorrection based on solar spectra cross-correlation.
    
    This class constructs a WavelengthOffset and applies the resulting correction
    from a cross-correlation between a reference spectrum of the Sun and a twilight
    exposure, dominated by solar spectra features.

    Attributes
    ----------
    sun_intensity : np.ndarray
        Reference solar spectrum.
    sun_wavelength : np.ndarray
        Wavelength vector associated to ``sun_intensity``
    """
    name = "SolarCrossCorrelationOffset"

    def __init__(self, sun_wavelength, sun_intensity, **kwargs):
        super().__init__(offset=WavelengthOffset(), **kwargs)
        self.sun_wavelength = check_unit(sun_wavelength, u.AA)
        self.sun_intensity = check_unit(sun_intensity,
                                        u.erg / u.s / u.AA / u.cm**2)

    @classmethod
    def from_fits(cls, path=None, extension=1):
        """Initialise a WavelegnthOffset correction from an input FITS file.
        
        Parameters
        ----------
        path : str, optional
            Path to the FITS file containing the reference Sun's spectra. The
            file must contain an extension with a table including a ``WAVELENGTH``
            and ``FLUX`` columns.The wavelength array must be angstrom in the
            vacuum frame.
        extension : int or str, optional
            HDU extension containing the table. Default is 1.

        Returns
        -------
        solar_offset_correction : :class:`SolarCrossCorrOffset`
            An instance of SolarCrossCorrOffset.
        """
        if path is None:
            path = os.path.join(os.path.dirname(__file__), '..',
                     'input_data', 'spectrophotometric_stars',
                     'sun_mod_001.fits')
        with fits.open(path) as hdul:
            sun_wavelength = hdul[extension].data['WAVELENGTH'] << u.AA
            sun_wavelength = vac_to_air(sun_wavelength)
            sun_intensity = hdul[extension].data['FLUX'] << u.erg / u.s / u.AA / u.cm**2
        return cls(sun_wavelength=sun_wavelength,
                   sun_intensity=sun_intensity)

    @classmethod
    def from_text_file(cls, path, loadtxt_args={}):
        """Initialise a :class:`SolarCrossCorrOffset` correction from an input text file.
        
        Parameters
        ----------
        path: str
            Path to the text file containing the reference Sun's spectra. The
            text file must contain two columns consisting of the
            vacuum wavelength array in angstrom and the solar flux or luminosity.
        loadtxt_args: dict, optional
            Additional arguments to be passed to ``numpy.loadtxt``.

        Returns
        -------
        solar_offset_correction: :class:`SolarCrossCorrOffset`
            An instance of SolarCrossCorrOffset.
        """
        sun_wavelength, sun_intensity = np.loadtxt(path, unpack=True,
                                                   usecols=(0, 1),
                                                   **loadtxt_args)
        #TODO: Handle units
        sun_wavelength = vac_to_air(sun_wavelength)
        return cls(sun_wavelength=sun_wavelength,
                   sun_intensity=sun_intensity)


    def get_solar_features(self, solar_wavelength, solar_spectra,
                            window_size_aa=20):
        """
        Estimate the regions of the solar spectrum dominated by absorption features.

        Description
        -----------
        First, a median filter is applied to estimate the upper envelope of the
        continuum. Then, the median ratio between the solar spectra and the median-filtered
        estimate is used to compute the relative weights:

        .. math::
            \\begin{equation}
                w = \\left\\|\\frac{F_\\odot}{F_{cont}} - Median(\\frac{F_\\odot}{F_{cont}})\\right\\|
            \\end{equation}

        Parameters
        ----------
        solar_wavelength: numpy.ndarray
            Solar spectra wavelengths array.
        solar_spectra: numpy.ndarray
            Array containing the flux of the solar spectra associated to a given
            wavelength.
        window_size_aa: int, optional
            Size of a spectral window in angstrom to perform a median filtering
            and estimate the underlying continuum. Default is 20 AA.

        Returns
        -------
        weights: numpy.ndarray
            Array of weights representing the prominance of an absorption feature.

        """
        self.vprint("Estimating regions of solar spectra dominated by absorption lines.")
        delta_pixel = int(
            (check_unit(window_size_aa, u.AA) / (solar_wavelength[-1] - solar_wavelength[0])
                          * solar_wavelength.size).decompose()
                          )
        if delta_pixel % 2 == 0:
            delta_pixel += 1
        solar_continuum = median_filter(solar_spectra, delta_pixel
                                        ) << solar_spectra.unit
        # Detect absorption features
        median_continuum_ratio = np.nanmedian(solar_spectra / solar_continuum)
        weights = np.abs(solar_spectra / solar_continuum -  median_continuum_ratio)
        return weights

    def compute_grid_of_models(self, pix_shift_array, pix_std_array, pix_array,
                              sun_intensity, weights):
        """Compute a grid of Solar spectra models convolved with a gaussian LSF.
        
        Parameters
        ----------
        pix_shift_array: 1D-np.array
            Array containing the wavelength offsets expressed in pixels.
        pix_std_array: 1D-np.array
            Array containing the values of the gaussian LSF standard deviation
            in pixels.
        pix_array: 1D-np.array
            Array of pixels to sample the grid of models.
        sun_intensity: 1D-np.array
            Array of solar fluxes associated to ``pix_array``.
        weights: 1D-np.array
            Array of absorption-features weights associated to ``sun_intensity``.
        
        Returns
        -------
        models_grid: numpy.ndarray
            Grid of models with dimensions `(n, m, s)`, where `n`, `m` and `s`
            correspond to the size of `pix_shift_array`, `pix_std_array`, and
            `pix_array`, respectively.
        weights_grid: numpy.ndarray
            Grid of absorption-feature weights associated to `models_grid`.

        See also
        --------
        :For more details on the computation of the weights array see :func:`get_solar_features`.

        """
        self.vprint("Computing grid of solar spectra models")
        models_grid = np.zeros((pix_shift_array.size, pix_std_array.size,
                                sun_intensity.size)) << sun_intensity.unit
        weights_grid = np.zeros((pix_shift_array.size, pix_std_array.size,
                                 sun_intensity.size))
        shift_idx, std_idx = np.indices(models_grid.shape[:-1])

        for z, (velshift, gauss_std) in enumerate(
            zip(pix_shift_array[shift_idx.flatten()],
                pix_std_array[std_idx.flatten()])):
                i, j = np.unravel_index(z, models_grid.shape[:-1])
                new_pixel_array = pix_array + velshift
                interp_sun_intensity = flux_conserving_interpolation(
                    new_pixel_array, pix_array, sun_intensity)
                interp_sun_intensity = gaussian_filter(
                    interp_sun_intensity, gauss_std.value)
                # Restore the intensity units removed by gaussian_filter
                models_grid[i, j] = interp_sun_intensity << sun_intensity.unit
                # Repeat the process for the weights
                interp_sun_weight = flux_conserving_interpolation(
                    new_pixel_array, pix_array, weights)
                # Truncate the gaussian filter to 2-sigma
                interp_sun_weight = gaussian_filter(
                    interp_sun_weight, gauss_std.value, truncate=2.0)
                interp_sun_weight /= np.nansum(interp_sun_weight)
                weights_grid[i, j] = interp_sun_weight
        return models_grid, weights_grid

    def compute_shift_from_twilight(self, spectra_container,
                                    sun_window_size_aa=20, keep_features_frac=0.1,
                                    response_window_size_aa=200,
                                    wave_range=None,
                                    pix_shift_array=None,
                                    pix_std_array=None,
                                    logspace=True, use_mean=True,
                                    inspect_fibres=None):
        """Compute the wavelenght offset of between a given SpectraContainer and a reference Solar spectra.
        
        Parameters
        ----------
        spectra_container: `pykoala.data_container.SpectraContainer`
            Spectra container (RSS or Cube) to cross-correlate with the reference
            spectra.
        sun_window_size_aa: int, optional
            Size of a spectral window in angstrom to perform a median filtering
            and estimate the underlying continuum. Default is 20 AA.
            See `get_solar_features` for details.
        keep_features_frac: float, optional
            Fraction of absorption-features weights to keep. All values below
            that threshold will be set to 0. Default is 0.1.
        wave_range: list or tuple, optional
            If provided, the cross-correlation will only be done in the provided
            wavelength range. Default is None.
        pix_shift_array: 1D-np.array, optional, default=np.arange(-5, 5, 0.1)
            Array containing the wavelength offsets expressed in pixels.
        pix_std_array: 1D-np.array, optional, default=np.arange(0.1, 3, 0.1)
            Array containing the values of the gaussian LSF standard deviation
            in pixels. See `compute_grid_of_models` for details.
        logspace: bool, optional
            If True, the cross-correlation will be perform using a logarithmic
            sampling in terms of wavelength. Default is True.
        use_mean: bool, optional
            If True, the mean likelihood-weighted value of the wavelength offset
            is used to create the `WavelengthOffsetCorrection`. Otherwise, the
            best fit parameters of the input grid are used. Default is True.
        inspect_fibres: list or tuple, optional
            Iterable containing RSS-wise spectra indices. If provided, a
            quality-control plot of each fibre is produced.
        
        Returns
        -------
        results: dict
            The dictionary contains the ``best-fit`` and ``mean`` likelihood-weighted
            values of ``pix_shift_array`` and ``pix_std_array`` in a tuple, respectively.
            If ``inspect_fibres`` is not ``None``, it containes a list of figures
            for each fibre included in ``inspect_fibres``.

        """
        if pix_shift_array is None:
            pix_shift_array = np.arange(-5, 5, 0.1)
        if pix_std_array is None:
            pix_std_array = np.arange(0.1, 3, 0.1)

        pix_shift_array = check_unit(pix_shift_array, u.pixel)
        pix_std_array = check_unit(pix_std_array, u.pixel)

        if logspace:
            new_wavelength = np.geomspace(spectra_container.wavelength[0],
                                          spectra_container.wavelength[-1],
                                          spectra_container.wavelength.size)
            rss_intensity = np.array([flux_conserving_interpolation(
                new_wavelength, spectra_container.wavelength, fibre
                ) for fibre in spectra_container.rss_intensity]
                ) << spectra_container.intensitu.unit
        else:
            new_wavelength = spectra_container.wavelength
            rss_intensity = spectra_container.rss_intensity
        
        # Interpolate the solar spectrum to the new grid of wavelengths
        sun_intensity = flux_conserving_interpolation(
        new_wavelength, self.sun_wavelength, self.sun_intensity)

        # Make an array of weights to focus on the absorption lines
        if wave_range is None:
            weights = self.get_solar_features(new_wavelength, sun_intensity,
                                            window_size_aa=sun_window_size_aa)
            weights[weights < np.nanpercentile(weights, 1 - keep_features_frac)] = 0
            weights[:100] = 0
            weights[-100:] = 0
        else:
            weights = np.zeros(new_wavelength)
            weights[slice(*np.searchsorted(new_wavelength, wave_range))] = 1.0

        valid_pixels = weights > 0
        self.vprint("Number of pixels with non-zero weights: "
                    + f"{np.count_nonzero(valid_pixels)} out of {valid_pixels.size}")

        # Estimate the response curve for each individual fibre
        delta_pixel = int(
            (check_unit(response_window_size_aa, u.AA)
                        / (new_wavelength[-1] - new_wavelength[0])
                        * new_wavelength.size).decompose()
                        )
        if delta_pixel % 2 == 0:
            delta_pixel += 1

        response_spectrograph = rss_intensity / sun_intensity[np.newaxis]
        smoothed_r_spectrograph = median_filter(
            response_spectrograph, delta_pixel, axes=1) << response_spectrograph.unit
        spectrograph_upper_env = percentile_filter(
            smoothed_r_spectrograph, 95, delta_pixel, axes=1) << response_spectrograph.unit
        # Avoid regions dominated by telluric absorption
        self.vprint("Including the masking of pixels dominated by telluric absorption")
        fibre_weights =  1 / (1  + (
                spectrograph_upper_env / smoothed_r_spectrograph
                - np.nanmedian(spectrograph_upper_env / smoothed_r_spectrograph)
                )**2)

        normalized_rss_intensity = rss_intensity / smoothed_r_spectrograph
        # Generate and fit the model
        pix_array = np.arange(new_wavelength.size) << u.pixel

        models_grid, weights_grid = self.compute_grid_of_models(
            pix_shift_array, pix_std_array, pix_array, sun_intensity, weights)

        # loop over one variable to avoir memory errors
        all_chi2 = np.zeros((pix_shift_array.size,
                             pix_std_array.size,
                             rss_intensity.shape[0]))
        
        self.vprint("Performing the cross-correlation with the grid of models")
        for i in range(pix_shift_array.size):
            all_chi2[i] = np.nansum(
                (models_grid[i, :, np.newaxis]
                 - normalized_rss_intensity[np.newaxis, :, :]).value**2
                * weights_grid[i, :, np.newaxis]
                * fibre_weights[np.newaxis, :, :],
                axis=-1) / np.nansum(
                    weights_grid[i, :, np.newaxis]
                    * fibre_weights[np.newaxis, :, :],
                    axis=-1)
            
        likelihood = np.exp(- (all_chi2 - all_chi2.min())/ 2)
        likelihood /= np.nansum(likelihood, axis=(0, 1))[np.newaxis, np.newaxis, :]
        mean_pix_shift = np.sum(likelihood.sum(axis=1)
                                * pix_shift_array[:, np.newaxis], axis=0)
        mean_std = np.sum(likelihood.sum(axis=0)
                          * pix_std_array[:, np.newaxis], axis=0)

        best_fit_idx = np.argmax(likelihood.reshape((-1, likelihood.shape[-1])),
                                 axis=0)
        best_vel_idx, best_std_idx = np.unravel_index(
                best_fit_idx, all_chi2.shape[:-1])
        best_sigma, best_shift = (pix_std_array[best_std_idx],
                                    pix_shift_array[best_vel_idx])

        if inspect_fibres is not None:
            fibre_figures = self.inspect_fibres(
                inspect_fibres, pix_shift_array, pix_std_array,
                best_vel_idx, best_std_idx, mean_pix_shift, mean_std,
                likelihood, models_grid, weights_grid, normalized_rss_intensity,
                new_wavelength)
        else:
            fibre_figures= None
        if use_mean:
            self.vprint("Using mean likelihood-weighted values to compute the wavelength offset correction")
            self.offset.offset_data = - mean_pix_shift
        else:
            self.vprint("Using best fit values to compute the wavelength offset correction")
            self.offset.offset_data = - best_shift
        
        self.offset.offset_error = np.full_like(best_shift, fill_value=np.nan)

        return {"best-fit": (best_shift, best_sigma),
                "mean": (mean_pix_shift, mean_std),
                "fibre_figures": fibre_figures}
    
    def inspect_fibres(self, fibres, pix_shift_array, pix_std_array,
                       best_vel_idx, best_std_idx, mean_vel, mean_std,
                       likelihood,
                       models_grid, weights_grid,
                       normalized_rss_intensity, wavelength):
        """
        Create a quality control plot of the solar cross-correlation process of each input fibre.

        Parameters
        ----------
        fibres: iterable
            List of input fibres to check.
        pix_shift_array: 1D-np.array
            Array containing the wavelength offsets expressed in pixels.
        pix_std_array: 1D-np.array
            Array containing the values of the gaussian LSF standard deviation
            in pixels. See :func:`compute_grid_of_models` for details.
        best_vel_idx: int
            Index of ``pix_shift_array`` that correspond to the best fit.
        best_std_idx: int
            Index of ``pix_std_array`` that correspond to the best fit.
        mean_vel: float
            Mean likelihood-weighted values of ``pix_shift_array``.
        mean_std: float
            Mean likelihood-weighted values of ``pix_std_array``.
        likelihood: numpy.ndarray:
            Likelihood of the cross-correlation.
        models_grid: numpy.ndarray
            Grid of solar spectra models. See :func:`compute_grid_of_models` for details.
        weights_grid: numpy.ndarray
            Grid of solar spectra weights. See :func:`compute_grid_of_models` for details.
        normalized_rss_intensity: numpy.ndarray
            Array containing the RSS intensity values of a SpectraContainer including
            the correction of the spectrograph response curve.
        wavelength: np.array
            Wavelength array associated to ``normalized_rss_intensity`` and ``models_grid``.

        Returns
        -------
        fibres_figures: list
            List of figures containing a QC plot of each fibre.
        """
        fibre_figures = []
        best_sigma, best_shift = (pix_std_array[best_std_idx],
                                  pix_shift_array[best_vel_idx])
        for fibre in fibres:
            self.vprint(f"Inspecting input fibre: {fibre}")
            fig = plt.figure(constrained_layout=True, figsize=(10, 8))
            gs = GridSpec(2, 4, figure=fig, wspace=0.25, hspace=0.25)

            ax = fig.add_subplot(gs[0, 0])
            mappable = ax.pcolormesh(
                pix_std_array, pix_shift_array, likelihood[:, :, fibre],
                cmap='gnuplot',
                norm=LogNorm(vmin=likelihood.max() / 1e2, vmax=likelihood.max()))
            plt.colorbar(mappable, ax=ax,
                         label=r"$e^(-\sum_\lambda w(I - \hat{I}(s, \sigma))^2 / 2)$")
            ax.plot(best_sigma[fibre], best_shift[fibre], '+', color='cyan',
                    label=r'Best fit: $\Delta\lambda$='
                    + f'{best_shift[fibre]:.2}, ' + r'$\sigma$=' + f'{best_sigma[fibre]:.2f}')
            ax.plot(mean_std[fibre], mean_vel[fibre], 'o', mec='lime', mfc='none',
                    label=r'Mean value: $\Delta\lambda$='
                    + f'{mean_vel[fibre]:.2}, ' + r'$\sigma$=' + f'{mean_std[fibre]:.2f}')
            ax.set_xlabel(r"$\sigma$ (pix)")
            ax.set_ylabel(r"$\Delta \lambda$ (pix)")
            ax.legend(bbox_to_anchor=(0., 1.05), loc='lower left', fontsize=7)

            sun_intensity = models_grid[best_vel_idx[fibre],
                                        best_std_idx[fibre]]
            weight = weights_grid[best_vel_idx[fibre],
                                        best_std_idx[fibre]]

            ax = fig.add_subplot(gs[0, 1:])
            ax.set_title(f"Fibre: {fibre}")
            ax.plot(wavelength, sun_intensity, label='Sun Model')
            ax.plot(wavelength, normalized_rss_intensity[fibre],
                    label='Fibre', lw=2)
            twax = ax.twinx()
            twax.plot(wavelength, weight, c='fuchsia',
                    zorder=-1, alpha=0.5, label='Weight')
            ax.legend(fontsize=7)
            ax.set_ylabel("Intensity")
            ax.set_xlabel("Wavelength")
            twax.set_ylabel("Relative weight")

            ax = fig.add_subplot(gs[1, :])
            max_idx = np.argmax(weight)
            max_weight_range = range(np.max((max_idx - 80, 0)),
                  np.min((max_idx + 80, wavelength.size - 1)))
            ax.plot(wavelength[max_weight_range],
                    sun_intensity[max_weight_range], label='Model')
            ax.plot(wavelength[max_weight_range],
                    normalized_rss_intensity[fibre][max_weight_range],
                    label='Fibre', lw=2)
            
            ax.set_xlim(wavelength[max_weight_range][0],
                        wavelength[max_weight_range][-1])
            twax = ax.twinx()
            twax.plot(wavelength[max_weight_range], weight[max_weight_range],
                      c='fuchsia',
                    zorder=-1, alpha=0.5, label='Absorption-feature Weight')
            twax.axhline(0)
            twax.legend(fontsize=7)
            ax.set_ylabel("Intensity")
            ax.set_xlabel("Wavelength")
            twax.set_ylabel("Relative weight")
            fibre_figures.append(fig)
            plt.close(fig)
        return fibre_figures


# =============================================================================
# Mr Krtxo \(ﾟ▽ﾟ)/
#                                                       ... Paranoy@ Rulz! ;^D



"""
This module contains the currenly available IFS intruments that `pykoala` can
ingest. Each submodule includes basic routines for reading and formating both data
and metadata from instrument-native files.
"""

from . import koala_ifu
#from . import hector_ifu
from . import weave

__all__ = ["koala_ifu"]


"""
This module contains the main tools for building 3D datacubes by interpolating
RSS data.
"""
# =============================================================================
# Basics packages
# =============================================================================
from abc import abstractmethod
import numpy as np
from scipy.special import erf

# =============================================================================
# Astropy and associated packages
# =============================================================================
from astropy.wcs import WCS
from astropy.coordinates import SkyCoord
from astropy import units as u

# =============================================================================
# KOALA packages
# =============================================================================
from pykoala import ancillary
from pykoala.data_container import Cube, RSS
from pykoala.plotting.utils import qc_cubing, qc_fibres_on_fov, qc_cube
from pykoala import vprint, VerboseMixin


class CubeStacking:
    """Collection of cubing stacking methods.

    Each method takes as input arguments a collection of cubes and variances,
    either in the form of a list or as an array with the first dimension corresponding
    to each cube, and additional keyword arguments.
    """

    @staticmethod
    def sigma_clipping(cubes: np.ndarray, variances: np.ndarray, **kwargs):
        """Perform cube stacking using STD clipping.

        Parameters
        ----------
        cubes: np.ndarray
            An array consisting of the collection of data to combine. The first
            dimension must correspond to the individual elements (e.g. datacubes)
            that will be combined. If the size of the first dimension is 1, it
            will return `cubes[0]` withouth applying any combination.
        variances: np.ndarray
            Array of variances associated to cubes.
        inv_var_weight: np.ndarray, optional
            An array of weights to apply during the stacking.

        Returns
        -------
        stacked_cube: np.ndarray
            The result of stacking the data in cubes along axis 0.
        stacked_variance: np.ndarray
            The result of stacking the variances along axis 0.
        """
        if cubes.shape[0] == 1:
            vprint("Only one cube to stack")
            return cubes[0], variances[0]

        nsigma = kwargs.get("nsigma", 3.0)
        sigma = np.nanstd(cubes, axis=0)
        mean = np.nanmean(cubes, axis=0)

        good_pixel = np.abs((cubes - mean[np.newaxis]) / sigma[np.newaxis]) < nsigma
        w = np.where(good_pixel, 1.0, 0.0)
        if kwargs.get("inv_var_weight", False):
            w = np.where(
                (variances > 0) & np.isfinite(variances) & good_pixel,
                1 / variances,
                np.nan,
            )
        norm = np.nansum(w, axis=0)
        illuminated_spx = norm > 0
        w = np.where(illuminated_spx, w / norm, np.nan)
        n_pix = np.sum(good_pixel, axis=0)
        stacked_cube = np.nansum(cubes * w, axis=0)
        # Do not include the pixels that are flagged as bad
        stacked_variance = np.full_like(stacked_cube, fill_value=np.nan)
        stacked_variance = np.where(
            n_pix > 0, np.nansum(variances * good_pixel, axis=0) / n_pix**2, np.nan
        )
        return stacked_cube, stacked_variance

    @staticmethod
    def mad_clipping(cubes: np.ndarray, variances: np.ndarray, **kwargs):
        """Perform cube stacking using MAD clipping.

        Parameters
        ----------
        cubes: np.ndarray
            An array consisting of the collection of data to combine. The first
            dimension must correspond to the individual elements (e.g. datacubes)
            that will be combined. If the size of the first dimension is 1, it
            will return `cubes[0]` withouth applying any combination.
        variances: np.ndarray
            Array of variances associated to cubes.
        inv_var_weight: np.ndarray, optional
            An array of weights to apply during the stacking.

        Returns
        -------
        stacked_cube: np.ndarray
            The result of stacking the data in cubes along axis 0.
        stacked_variance: np.ndarray
            The result of stacking the variances along axis 0.
        """
        if cubes.shape[0] == 1:
            vprint("Only one cube to stack")
            return cubes[0], variances[0]

        nsigma = kwargs.get("nsigma", 3.0)
        sigma = ancillary.std_from_mad(cubes, axis=0)
        median = np.nanmedian(cubes, axis=0)
        # Spaxel weights
        good_pixel = np.abs((cubes - median[np.newaxis]) / sigma[np.newaxis]) < nsigma
        w = np.where(good_pixel, 1.0, 0.0)
        if kwargs.get("inv_var_weight", False):
            w = np.where(
                (variances > 0) & np.isfinite(variances) & good_pixel,
                1 / variances,
                np.nan,
            )
        # Renormalize the weights using only pixels with data
        norm = np.nansum(w, axis=0)
        w = np.where(norm > 0, w / norm, np.nan)
        stacked_cube = np.nansum(cubes * w, axis=0)
        # Do not include the pixels that are flagged as bad
        stacked_variance = np.full_like(stacked_cube, fill_value=np.nan)
        n_pix = np.sum(good_pixel, axis=0)
        stacked_variance = np.where(
            n_pix > 0, np.nansum(variances * good_pixel, axis=0) / n_pix**2, np.nan
        )
        return stacked_cube, stacked_variance


# -------------------------------------------
# Fibre Interpolation and cube reconstruction
# -------------------------------------------


class InterpolationKernel(object):
    r"""Interpolation Kernel.

    A Kernel is a window function, :math:`K(u)`, used to perform the
    interpolation of individual fibre spectra at a given location in the sky
    :math:`(\alpha_0,\,\delta_0)` into a 3D grid.

    .. math::

        I(\alpha, \delta) = \int_{-\inf}^{\inf} I_{fib}(\alpha_0,\,\delta_0) \cdot K(\alpha - \alpha_0, \delta - \delta_0) d\alpha d\delta

    Different kernels have different domains, e.g. :math:`u\leq1` for
    Parabolic kernel, and therefore a scale parameter is sometimes required to
    renormalize the distance.

    The `truncation_radius`, expressed in units of `u`, is used with kernels
    whose domain extends all real numbers (e.g. Gaussian).
    """

    @property
    def scale(self) -> u.Quantity:
        """Kernel scale size in pixels."""
        return self._scale

    @scale.setter
    def scale(self, scale):
        self._scale = ancillary.check_unit(
            scale, u.pixel, equivalencies=self.pixel_scale
        )

    @property
    def scale_arcsec(self) -> u.Quantity:
        """Kernel scale size in arcseconds."""
        return self.scale.to(u.arcsec, self.pixel_scale)

    @property
    def truncation_radius(self) -> float:
        """Maximum value of `u` beyond which the kernel is set to 0."""
        return self._truncation_radius

    @truncation_radius.setter
    def truncation_radius(self, value):
        self._truncation_radius = value

    @property
    def pixel_scale(self) -> u.Equivalency:
        """Target pixel scale in arcsec per pixel."""
        return self._pixel_scale

    @pixel_scale.setter
    @u.quantity_input(pixel_scale=u.arcsec / u.pixel)
    def pixel_scale(self, pixel_scale):
        self._pixel_scale = u.pixel_scale(pixel_scale)

    def __init__(self, pixel_scale, scale, **kwargs):
        self.pixel_scale = pixel_scale
        self.scale = scale
        self.truncation_radius = kwargs.get("truncation_radius", 1.0)

    @abstractmethod
    def kernel_1D(self):
        pass

    @abstractmethod
    def kernel_2D(self):
        pass


class ParabolicKernel(InterpolationKernel):
    r"""Parabolic or Epanechnikov InterpolationKernel.

    The parabolic kernel is defined as:

    .. math::

        K(u) = \frac{3}{4}(1 - u^2),\, u\in [-1, 1]

    This model enforces `truncation_radius=1`.
    """

    def __init__(self, pixel_scale, scale, **kwargs):
        if "truncation_radius" in kwargs:
            del kwargs["truncation_radius"]
        super().__init__(pixel_scale, scale, truncation_radius=1, **kwargs)

    def cmf(self, ker_u):
        """Cumulative mass distribution.

        Parameters
        ----------
        ker_u : np.ndarray
            Positions in units of the kernel scale.

        Returns
        -------
        cmf : np.ndarray
            The values of the cumulative distribution evaluated at the input
            values of ker_u.
        """
        ker_u_clip = np.atleast_1d(ker_u).clip(-1, 1)
        return (3.0 * ker_u_clip - ker_u_clip**3 + 2.0) / 4

    @u.quantity_input(x_edges=u.pixel)
    def kernel_1D(self, x_edges):
        """Compute the kernel weights associated to a 1D array of bins.

        Parameters
        ----------
        x_edges : np.ndarray
            Array of values defining the edges of the bins as
            `x_bin=[x_edges[i], x_edges[i+1]]`. The values must be increasing.

        Returns
        -------
        weights : np.ndarray
            Array of kernel weights within each bin. The size is `len(x_edges) - 1`.
        """
        # Check that the input values are in increasing order
        assert (
            x_edges[1:] > x_edges[:-1]
        ).all(), "Input x_edges values must be increasing"
        # Convert pixels to kernel scale units
        u_edges = (x_edges / self.scale).clip(-1, 1)
        # Compute the cumulative distribution
        cumulative = self.cmf(u_edges)
        # Compute the weight on each bin
        weights = np.diff(cumulative)
        return weights

    @u.quantity_input(x_edges=u.pixel, y_edges=u.pixel)
    def kernel_2D(self, x_edges, y_edges):
        """Compute the kernel weights defined by bin edges in x and y directions.

        Parameters
        ----------
        x_edges : np.ndarray
            Array defining the bin edges along the x axis.
        y_edges : np.ndarray
            Array defining the bin edges aling the y axis.

        Returns
        -------
        weights : np.ndarray
            Array of kernel weights within each bin. The size is
            `(len(y_edges) - 1, len(x_edges) - 1)`.
        """
        grid_yy, grid_xx = np.meshgrid(
            y_edges / self.scale, x_edges / self.scale, indexing="ij"
        )
        cum_k = self.cmf(grid_xx) * self.cmf(grid_yy)
        weights = np.diff(cum_k, axis=0)
        weights = np.diff(weights, axis=1)
        return weights


class GaussianKernel(InterpolationKernel):
    r"""Gaussian InterpolationKernel.

    The Gaussian kernel is defined as:

    .. math::

        K(u) = \frac{1}{\sqrt{2\pi}} e^{-u^2/2},\, u\in [-\infty, \infty]

    The kernel domain is restricted to the range ``[-truncation_radius, truncation_radius]``.
    By default ``truncation_radius=3``.
    """

    def __init__(self, pixel_scale, scale, truncation_radius=3.0, **kwargs):
        super().__init__(
            pixel_scale, scale, truncation_radius=truncation_radius, **kwargs
        )
        # Minimum and maximum percentiles used for renormalizing the kernel
        self.left_norm = 0.5 * (1 + erf(-self.truncation_radius / np.sqrt(2)))
        self.right_norm = 0.5 * (1 + erf(self.truncation_radius / np.sqrt(2)))

    def cmf(self, ker_u):
        """Kernel cumulative distribution function.

        Parameters
        ----------
        ker_u : np.ndarray
            Positions in units of the kernel scale.

        Returns
        -------
        cmf : np.ndarray
            The values of the cumulative distribution evaluated at the input
            values of ker_u.
        """
        cmf = (0.5 * (1 + erf(ker_u / np.sqrt(2))) - self.left_norm) / (
            self.right_norm - self.left_norm
        )
        return cmf.clip(0, 1)

    @u.quantity_input(x_edges=u.pixel)
    def kernel_1D(self, x_edges, axis=0):
        cumulative = self.cmf(x_edges / self.scale)
        weights = np.diff(cumulative, axis=axis)
        return weights

    def kernel_2D(self, x_edges, y_edges):
        weights = (
            self.kernel_1D(x_edges)[np.newaxis, :]
            * self.kernel_1D(y_edges)[:, np.newaxis]
        )
        return weights


class TopHatKernel(InterpolationKernel):
    r"""TopHat (uniform) InterpolationKernel.

    The TopHat kernel is defined as:

    .. math::

        K(u) = \frac{1}{2},\, u\in [-1, 1]

    With :math:`u` restricted to the range [-1, 1] (i.e., this model enforces
    `truncation_radius=1`).
    """

    def __init__(self, pixel_scale, scale, **kwargs):
        if "truncation_radius" in kwargs:
            del kwargs["truncation_radius"]
        super().__init__(pixel_scale, scale, truncation_radius=1, **kwargs)

    def cmf(self, ker_u):
        """Kernel cumulative distribution function.

        Parameters
        ----------
        ker_u : np.ndarray
            Positions in units of the kernel scale.

        Returns
        -------
        cmf : np.ndarray
            The values of the cumulative distribution evaluated at the input
            values of ker_u.
        """
        cmf = 0.5 * (ker_u + 1)
        return cmf.clip(0, 1)

    @u.quantity_input(x_edges=u.pixel)
    def kernel_1D(self, x_edges, axis=0):
        cumulative = self.cmf(x_edges / self.scale)
        weights = np.diff(cumulative, axis=axis)
        return weights

    def kernel_2D(self, x_edges, y_edges):
        weights = (
            self.kernel_1D(x_edges)[np.newaxis, :]
            * self.kernel_1D(y_edges)[:, np.newaxis]
        )
        return weights


class CircularTopHatKernel(TopHatKernel):
    """Circular TopHat kernel."""

    def kernel_2D(self, x_edges, y_edges):
        rad = x_edges[np.newaxis, :] ** 2 + y_edges[:, np.newaxis] ** 2
        return self.kernel_1D(rad, axis=(0, 1))


class DrizzlingKernel(TopHatKernel):
    """Drizzling InterpolationKernel.

    This kernel follows the same phylosophy as in Fruchter & Hook 1997.
    Users may define different fibre footprint shapes (``fibre_shape``) among:
    ``squared``, ``circular``, or ``hexagonal``. This will determine what is the
    overlapping fraction between a given spaxel and the fibre.
    """

    fibre_frac = {
        "circle": ancillary.pixel_in_circle,
        "hexagon": ancillary.pixel_in_hexagon,
        "square": ancillary.pixel_in_square,
    }
    """Dictionary mapping the available fibre shapes to the corresponding method."""

    def __init__(self, pixel_scale, scale, **kwargs):
        super().__init__(pixel_scale, scale, **kwargs)
        # Assume circular fibre footprint by default
        self.pixel_frac_method = self.fibre_frac[kwargs.get("fibre_shape", "circle")]

    def kernel_1D(self):
        raise NotImplementedError("kernel_1D has not been implemented in this class")

    def kernel_2D(self, x_edges, y_edges):
        weights = np.zeros((y_edges.size - 1, x_edges.size - 1))
        # y == rows, x == columns
        # Only the lower-left corner of every pixel is required.
        pix_edge_y, pix_edge_x = np.meshgrid(
            y_edges[:-1].to_value("pixel"),
            x_edges[:-1].to_value("pixel"),
            indexing="ij",
        )
        for i, pos in enumerate(zip(pix_edge_x.flatten(), pix_edge_y.flatten())):
            _, area_fraction = self.pixel_frac_method(
                pixel_pos=pos,
                pixel_size=1,
                pos=(0.0, 0.0),
                radius=self.scale.to_value("pixel") / 2,
            )
            weights[np.unravel_index(i, weights.shape)] = area_fraction
        return weights


# ------------------------------------------------------------------------------
# Fibre interpolation
# ------------------------------------------------------------------------------


class CubeInterpolator(VerboseMixin):
    """A class for combining multiple RSS into a 3D datacube.

    This class performs the combination of a set of RSS data into a 3D regular
    grid defined by a WCS.

    The basic ingredients for creating a datacube are:
    - Set of row-stacked spectra (:class:`RSS`).
    - A target :class:`WCS` that defines the final datacube dimensions.
    - An :class:`InterpolationKernel` that will determine the contribution of
    each RSS at every location.

    In addition, users may provide additional information such as:
    - A set of differential atmospheric corrections (DAR) for each RSS that
    accounts for wavelength-dependent fibre spatial variations.
    - A set of flags of the RSS masks that will be used to mask out pixels during
    the interpolation.

    To assess the quality of the interpolation, users may want to keep intermediate
    products or create plots that allow to evaluate the performance of the cubing.
    First, the variables ``all_datacubes`` and ``all_var`` store the interpolated
    fluxes and variances of each individual RSS, respectively.
    The kernel weights and net exposure times are stored in the variables
    ``all_weights`` and ``all_exp_time``.
    Furthermore, if "keep_individual_cubes=True" the variable ``rss_inter_products``
    will store each individual :class:`Cube` for each RSS and a QC plot.

    Users may set the argument ``qc_plots=True``, that will create


    Example
    -------
    >>> from pykoala.instruments.mock import mock_rss
    >>> interpolator = CubeInterpolator(rss_set=[mock_rss()])
    >>> cube = interpolator.build_cube()
    """

    @property
    def rss_set(self) -> list:
        """List of target :class:`RSS` to be combined into a :class:`Cube`."""
        return self._rss_set
    
    @rss_set.setter
    def rss_set(self, rss_set):
        self._rss_set = rss_set
    
    @property
    def adr_set(self) -> list:
        """List of ADR to apply during interpolation."""
        return self._adr_set

    @adr_set.setter
    def adr_set(self, adr_set):
        self._adr_set = adr_set

    @property
    def mask_flags(self) -> list:
        """List of flags to be included during the pixel masking."""
        return self._mask_flags

    @mask_flags.setter
    def mask_flags(self, mask_flags):
        self._mask_flags = mask_flags

    @property
    def target_wcs(self) -> WCS:
        """Target WCS into which the RSS data will be interpolated."""
        return self._target_wcs
    
    @target_wcs.setter
    def target_wcs(self, wcs):
        self._target_wcs = wcs
    
    @property
    def kernel(self) -> InterpolationKernel:
        """:class:`InterpolationKernel` used to combine the RSS data."""
        return self._kernel

    @kernel.setter
    def kernel(self, kernel):
        self._kernel = kernel

    def __init__(
        self,
        rss_set,
        wcs=None,
        kernel=GaussianKernel,
        kernel_scale=2.0 << u.arcsec,
        kernel_truncation_radius=3.0,
        adr_set=None,
        mask_flags=None,
        qc_plots=False,
        **kwargs,
    ):
        # Verbosity parameters
        self.logger = kwargs.get("logger", "CubeInterp")
        self.verbose = kwargs.get("verbose", True)
        # Set of RSS to be interpolated
        self.rss_set = rss_set
        # Differential atmospheric refraction
        if adr_set is None:
            self.adr_set = [(None, None)] * len(rss_set)
        else:
            self.adr_set = adr_set
        # Names of flags to be used for masking pixels
        self.mask_flags = mask_flags
        # WCS defining the target dimensions of the cube
        if wcs is None:
            self.vprint("Computing WCS using input list of RSS")
            self.target_wcs = build_wcs_from_rss(
                self.rss_set,
                kwargs.get("spatial_pix_size", 1 << u.arcsec),
                kwargs.get("spectra_pix_size", 1 << u.AA),
            )
        else:
            self.vprint("Using input WCS")
            self.target_wcs = wcs
        # RSS Interpolation kernel
        if isinstance(kernel, type):
            kernel_scale = ancillary.check_unit(kernel_scale, u.arcsec)
            # Assume square pixel
            pixel_scale = (
                np.abs(self.target_wcs.celestial.pixel_scale_matrix.diagonal()).mean()
                << u.deg / u.pixel
            )

            self.vprint(
                f"Initialising {kernel.__name__}"
                + f"\n Scale: {kernel_scale:.1f} (pixels)"
                + f"\n Truncation radius: {kernel_truncation_radius:.1f}"
                + f" ({kernel_scale * kernel_truncation_radius:.1f} px)"
            )
            self.kernel = kernel(
                pixel_scale=pixel_scale,
                scale=kernel_scale,
                truncation_radius=kernel_truncation_radius,
            )
        else:
            self.kernel = kernel
            kernel_scale = self.kernel.scale_arcsec
            self.vprint(
                f"User-provided interpolation kernel"
                + f"\n Scale: {self.kernel.scale:.1f} (pixels)"
                + f"\n Truncation radius: {self.kernel.truncation_radius:.1f}"
            )
        # Create the output data units
        if u.second in rss_set[0].intensity.unit.bases:
            output_unit = rss_set[0].intensity.unit
        else:
            output_unit = rss_set[0].intensity.unit / u.second
        self.vprint(
            f"Initialising new Cube with dimensions: {self.target_wcs.array_shape}"
        )
        self.vprint(f"Output Cube units: {output_unit.to_string()}")

        self.all_datacubes = (
            np.full(
                (len(self.rss_set), *self.target_wcs.array_shape), fill_value=np.nan
            )
            << output_unit
        )
        self.all_var = (
            np.full(self.all_datacubes.shape, fill_value=np.nan) << output_unit**2
        )
        # Total weight per spaxel
        self.all_weights = np.full(self.all_datacubes.shape, fill_value=np.nan)
        # Exposure time per spaxel
        self.all_exp_time = (
            np.full(self.all_datacubes.shape, fill_value=np.nan) << u.second
        )
        # "Empty" array that will be used to store exposure times
        self.exposure_times = (
            np.array([rss.info["exptime"].to_value("second") for rss in self.rss_set])
            << u.second
        )

        # Create variables to store plots and intermediate products
        self.make_qc_plots = qc_plots
        if self.make_qc_plots:
            self.vprint("QC plots will be generated during cube production")
        self.keep_individual_cubes = kwargs.get("keep_individual_cubes", False)
        if self.keep_individual_cubes:
            self.vprint("Cubes for each individual RSS will be built")
        self.rss_inter_products = {}
        self.cube_plots = {}

    def _fill_info(self, info):
        """Fill the info metadata using the RSS set.
        
        Parameters
        ----------
        info : dict
            Dictionary containing the Cube metadata.
        
        Returns
        -------
        info : dict
            Updated dictionary
        """
        if "airmass" not in info:
            info["airmass"] = np.nanmean(
                [rss.info.get("airmass", np.nan) for rss in self.rss_set])
        if "exptime" not in info:
            info["exptime"] = np.nansum(
                [rss.info.get("exptime").to_value("s") for rss in self.rss_set
                 if rss.info.get("exptime") is not None]) << u.second
        return info

    def build_cube(
        self,
        stacking_method=CubeStacking.mad_clipping,
        stacking_args=None,
        cube_info={},
    ) -> Cube:
        """Perform the interpolation of the RSS into a  :class:`Cube`.

        Parameters
        ----------
        stacking_method : func, optional
            Stacking method to use for combining the RSS (See :class:`CubeStacking`).
        stacking_args : dict, optional
            Additional arguments passed to the stacking method.
        cube_info : dict, optional
            Additional metadata to be included in the final :class:`Cube` or
            the intermediate RSS products.

        Returns
        -------
        cube : :class:`Cube`
            The resulting cube from combining the input RSS.
        """
        self.vprint("Cubing input RSS set")

        # Save intermediate products
        for ith, rss in enumerate(self.rss_set):
            # Interpolate RSS to data cube
            (
                datacube_i,
                datacube_var_i,
                datacube_weight_i,
                interp_info,
            ) = self.interpolate_rss(
                rss,
                # Initialise the empty variables
                datacube=np.zeros(self.target_wcs.array_shape) << rss.intensity.unit,
                datacube_var=np.zeros(self.target_wcs.array_shape) << rss.variance.unit,
                datacube_weight=np.zeros(self.target_wcs.array_shape),
                # Differential Atmospheric Refraction
                adr_ra_arcsec=self.adr_set[ith][0],
                adr_dec_arcsec=self.adr_set[ith][1],
            )

            zero_weight = datacube_weight_i == 0
            datacube_i[zero_weight] = np.nan
            datacube_var_i[zero_weight] = np.nan
            datacube_weight_i[zero_weight] = np.nan

            self.all_weights[ith] = datacube_weight_i
            self.all_exp_time[ith] = datacube_weight_i * self.exposure_times[ith]

            if u.second in rss.intensity.unit.bases:
                self.all_datacubes[ith] = datacube_i / self.all_weights[ith]
                self.all_var[ith] = datacube_var_i / self.all_weights[ith] ** 2
            else:
                self.all_datacubes[ith] = datacube_i / self.all_exp_time[ith]
                self.all_var[ith] = datacube_var_i / self.all_exp_time[ith] ** 2

            # Create a single-RSS cube
            if self.keep_individual_cubes:
                ind_cube = Cube(
                    intensity=self.all_datacubes[ith],
                    variance=self.all_var[ith],
                    wcs=self.target_wcs,
                    info=dict(kernel_scale=self.kernel.scale_arcsec, name=f"rss_{ith}"),
                )
                interp_info["cube"] = (ind_cube, qc_cube(ind_cube))

            self.rss_inter_products[f"rss_{ith}"] = interp_info

        # Combine all cubes
        self.vprint(f"Stacking individual cubes using {stacking_method.__name__}")
        if stacking_args:
            self.vprint(f"Additonal arguments for stacking: {stacking_args}")
        else:
            stacking_args = {}
        datacube, datacube_var = stacking_method(
            self.all_datacubes, self.all_var, **stacking_args
        )
        info = dict(kernel_scale=self.kernel.scale_arcsec, **cube_info)
        info = self._fill_info(info)
        # Create the Cube
        cube = Cube(intensity=datacube, variance=datacube_var,
                    wcs=self.target_wcs, info=info)
        if self.make_qc_plots:
            self.vprint("Producing quality assessment plots")
            # Fibre coverage and exposure time maps
            self.cube_plots["weights"] = qc_cubing(self.all_weights, self.all_exp_time)
            # QC cube maps
            self.cube_plots["stack_cube"] = qc_cube(cube)
        return cube

    def interpolate_rss(
        self,
        rss,
        datacube,
        datacube_var,
        datacube_weight,
        adr_ra_arcsec=None,
        adr_dec_arcsec=None,
    ):
        """Perform fibre interpolation using a RSS into to a 3D datacube.

        Parameters
        ----------
        rss : :class:`RSS`
            Target RSS to be interpolated.
        datacube : u.Quantity, optional
            Array that stores the intensity associated to a datacube.
        datacube_var : u.Quantity, optional
            Array that stores the variance associated of a datacube.
        datacube_weights : np.ndarray, optional
            Array that stores the fibre weights of the datacube.
        adr_ra_arcsec : u.Quantity, optional
            Differential atmospheric refraction offset along the RA direction.
        adr_dec_arcsec : u.Quantity, optional
            Differential atmospheric refraction offset along the DEC direction.

        Returns
        -------
        datacube : u.Quantity
            Array contanining the interpolated intensity of the RSS into the
            target WCS.
        datacube_var : u.Quantity
            Array contanining the interpolated variance of the RSS into the
            target WCS.
        datacube_weight : u.Quantity
            Array contanining the sum of all kernels weights of all fibres.
        interm_products : dict
            Dictionary containing intermediate products such as individual
            fibre weights.
        """
        self.vprint("Interpolating RSS to cube")

        # Obtain fibre position in the detector (center of pixel)
        (
            fibre_pixel_pos_cols,
            fibre_pixel_pos_rows,
        ) = self.target_wcs.celestial.world_to_pixel(
            SkyCoord(rss.info["fib_ra"], rss.info["fib_dec"])
        )

        # Dictionary that stores additional information
        interm_products = {
            "fib_pix_col": fibre_pixel_pos_cols,
            "fib_pix_row": fibre_pixel_pos_rows,
        }

        # Wavelength array of target datacube
        cube_wavelength = self.target_wcs.spectral.array_index_to_world(
            np.arange(self.target_wcs.array_shape[0])
        ).to(rss.wavelength.unit)

        if adr_dec_arcsec is not None:
            adr_dec_pixel = adr_dec_arcsec.to(u.pixel, self.kernel.pixel_scale)
            if cube_wavelength.size != rss.wavelength.size or np.allclose(
                cube_wavelength, rss.wavelength
            ):
                adr_dec_pixel = np.interp(
                    cube_wavelength, rss.wavelength, adr_dec_pixel
                )
        else:
            adr_dec_pixel = None
        if adr_ra_arcsec is not None:
            adr_ra_pixel = adr_ra_arcsec.to(u.pixel, self.kernel.pixel_scale)
            if cube_wavelength.size != rss.wavelength.size or np.allclose(
                cube_wavelength, rss.wavelength
            ):
                adr_ra_pixel = np.interp(cube_wavelength, rss.wavelength, adr_ra_pixel)
        else:
            adr_ra_pixel = None
        # Interpolate all RSS fibres
        if self.mask_flags is not None:
            self.vprint(f"Flagging those pixels with: {self.mask_flags}")
            mask = rss.mask.get_flag_map(self.mask_flags)
        else:
            mask = np.zeros(rss.intensity.shape, dtype=bool)

        # Create a fibre coverage map
        if self.make_qc_plots:
            qc_fig = qc_fibres_on_fov(
                datacube.shape[1:],
                fibre_pixel_pos_cols,
                fibre_pixel_pos_rows,
                fibre_diam=getattr(rss, "fibre_diameter", 1.25 << u.arcsec).to(
                    u.pixel, self.kernel.pixel_scale
                ),
            )
            interm_products["qc_fibres_on_fov"] = qc_fig

        if cube_wavelength.size != rss.wavelength.size or np.allclose(
            cube_wavelength, rss.wavelength
        ):
            self.vprint("Fibres will be interpolated to new wavelength grid")
            interp_wave = True
        else:
            interp_wave = False
        interm_products["fibre_weights"] = []
        for fibre in range(rss.intensity.shape[0]):
            # Spectra interpolation
            if interp_wave:
                f_intensity = ancillary.flux_conserving_interpolation(
                    cube_wavelength, rss.wavelength, rss.intensity[fibre]
                )
                f_variance = ancillary.flux_conserving_interpolation(
                    cube_wavelength, rss.wavelength, rss.variance[fibre]
                )
                f_mask = np.interp(cube_wavelength, rss.wavelength, mask[fibre])
                # Mask all pixels contaminated
                f_mask = f_mask > 0
            else:
                f_intensity = rss.intensity[fibre]
                f_variance = rss.variance[fibre]
                f_mask = mask[fibre]

            # Interpolate fibre to cube
            datacube, datacube_var, datacube_weight = self.interpolate_fibre(
                fib_spectra=f_intensity,
                fib_variance=f_variance,
                cube=datacube,
                cube_var=datacube_var,
                cube_weight=datacube_weight,
                pix_pos_cols=fibre_pixel_pos_cols[fibre] << u.pixel,
                pix_pos_rows=fibre_pixel_pos_rows[fibre] << u.pixel,
                adr_cols=adr_ra_pixel,
                adr_rows=adr_dec_pixel,
                fibre_mask=f_mask,
                interm_products=interm_products,
            )
        return datacube, datacube_var, datacube_weight, interm_products

    def interpolate_fibre(
        self,
        fib_spectra,
        fib_variance,
        cube,
        cube_var,
        cube_weight,
        pix_pos_cols,
        pix_pos_rows,
        adr_cols=None,
        adr_rows=None,
        adr_pixel_frac=0.05,
        fibre_mask=None,
        interm_products=None,
    ):
        """Interpolates fibre spectra and variance to a 3D data cube.

        Parameters
        ----------
        fib_spectra: (k,) np.array(float)
            Array containing the fibre spectra.
        fib_variance: (k,) np.array(float)
            Array containing the fibre variance.
        cube: (k, n, m) np.ndarray (float)
            Cube to interpolate fibre spectra.
        cube_var: (k, n, m) np.ndarray (float)
            Cube to interpolate fibre variance.
        cube_weight: (k, n, m) np.ndarray (float)
            Cube to store fibre spectral weights.
        pix_pos_cols: int
            Fibre column pixel position (m).
        pix_pos_rows: int
            Fibre row pixel position (n).
        adr_cols: (k,) np.array(float), optional, default=None
            Atmospheric Differential Refraction (ADR) of each wavelength point
            along x (ra)-axis (m) expressed in pixels.
        adr_rows: (k,) np.array(float), optional, default=None
            Atmospheric Differential Refraction of each wavelength point along
            y (dec) -axis (n) expressed in pixels.
        adr_pixel_frac: float, optional, default=0.05
            ADR Pixel fraction used to bin the spectral pixels. For each bin,
            the median ADR correction will be used to
            correct the range of wavelength.
        fibre_mask: np.ndarray
            Boolean array containing the fibre mask.
        interm_products : dict
            Dictionary that stores intermediate products and metadata.

        Returns
        -------
        cube:
            Original datacube with the fibre data interpolated.
        cube_var:
            Original variance with the fibre data interpolated.
        cube_weight:
            Original datacube weights with the fibre data interpolated.
        interm_products : dict
            Dictionary that stores intermediate products and metadata.
        """
        if adr_rows is None and adr_cols is None:
            adr_rows = np.zeros(fib_spectra.size) << u.pixel
            adr_cols = np.zeros(fib_spectra.size) << u.pixel
            spectral_window = fib_spectra.size
        else:
            # Estimate spectral window
            adr_pixel_frac = ancillary.check_unit(adr_pixel_frac, u.pixel)
            spectral_window = int(
                np.min(
                    (
                        adr_pixel_frac / np.abs(adr_cols[0] - adr_cols[-1]),
                        adr_pixel_frac / np.abs(adr_rows[0] - adr_rows[-1]),
                    )
                )
                * fib_spectra.size
            )

        # Set NaNs to 0 and discard pixels
        if fibre_mask is None:
            fibre_mask = np.zeros(fib_spectra.shape, dtype=bool)
        nan_pixels = ~np.isfinite(fib_spectra) | fibre_mask
        fib_spectra[nan_pixels] = 0.0 << fib_spectra.unit

        pixel_weights = np.ones(fib_spectra.size)
        pixel_weights[nan_pixels] = 0.0

        # Loop over wavelength pixels
        fibre_weights = []
        for wl_range in range(0, fib_spectra.size, spectral_window):
            wl_slice = slice(wl_range, wl_range + spectral_window)
            # Kernel along columns direction (x, ra)
            kernel_centre_cols = pix_pos_cols - np.nanmedian(adr_cols[wl_slice])
            kernel_offset = self.kernel.scale * self.kernel.truncation_radius
            cols_min = max(int(kernel_centre_cols.value - kernel_offset.value) - 1, 0)
            cols_max = min(
                int(kernel_centre_cols.value + kernel_offset.value) + 1,
                cube.shape[2] - 1,
            )
            columns_slice = slice(cols_min, cols_max + 1, 1)
            # Kernel along rows direction (y, dec)
            kernel_centre_rows = pix_pos_rows - np.nanmedian(adr_rows[wl_slice])
            rows_min = max(int(kernel_centre_rows.value - kernel_offset.value) - 1, 0)
            rows_max = min(
                int(kernel_centre_rows.value + kernel_offset.value) + 1,
                cube.shape[1] - 1,
            )
            rows_slice = slice(rows_min, rows_max + 1, 1)

            if (cols_max < cols_min) | (rows_max < rows_min):
                continue

            # Compute the kernel weight associated to each location
            weights = self.kernel.kernel_2D(
                np.arange(cols_min - 0.5, cols_max + 1.5, 1.0) * u.pixel
                - kernel_centre_cols,
                np.arange(rows_min - 0.5, rows_max + 1.5, 1.0) * u.pixel
                - kernel_centre_rows,
            )

            fibre_weights.append((wl_slice, rows_slice, columns_slice, weights))

            weights = weights[np.newaxis]
            # Add spectra to cube
            cube[wl_slice, rows_slice, columns_slice] = np.add(
                cube[wl_slice, rows_slice, columns_slice],
                fib_spectra[wl_slice, np.newaxis, np.newaxis] * weights,
            )
            cube_var[wl_slice, rows_slice, columns_slice] = np.add(
                cube_var[wl_slice, rows_slice, columns_slice],
                fib_variance[wl_slice, np.newaxis, np.newaxis] * weights**2,
            )
            cube_weight[wl_slice, rows_slice, columns_slice] = np.add(
                cube_weight[wl_slice, rows_slice, columns_slice],
                pixel_weights[wl_slice, np.newaxis, np.newaxis] * weights,
            )

        interm_products["fibre_weights"].append(fibre_weights)
        return cube, cube_var, cube_weight


def build_wcs(
    datacube_shape,
    reference_position,
    spatial_pix_size: u.Quantity,
    spectra_pix_size: u.Quantity,
    radesys="ICRS    ",
    equinox=2000.0,
) -> WCS:
    """Create a WCS using cubing information.

    Integer pixel values fall at the center of pixels.

    Parameters
    ----------
    datacube_shape: tuple
        Pixel shape of the datacube (wavelength, ra, dec).
    reference_position : tuple
        Values corresponding to the origin of the wavelength axis, and sky position of the central pixel.
    spatial_pix_size : u.Quantity
        Pixel size along the spatial direction.
    spectra_pix_size : u.Quantity
        Pixel size along the spectral direction.

    """
    wcs_dict = {
        # Spatial dimensions
        "RADECSYS": radesys,
        "EQUINOX": equinox,
        "CTYPE1": "RA---TAN",
        "CUNIT1": "deg",
        "CDELT1": spatial_pix_size.to_value("deg"),
        "CRPIX1": datacube_shape[1] / 2,
        "CRVAL1": reference_position[1].to_value("deg"),
        "NAXIS1": datacube_shape[1],
        "CTYPE2": "DEC--TAN",
        "CUNIT2": "deg",
        "CDELT2": spatial_pix_size.to_value("deg"),
        "CRPIX2": datacube_shape[2] / 2,
        "CRVAL2": reference_position[2].to_value("deg"),
        "NAXIS2": datacube_shape[2],
        # Spectral dimension
        "CTYPE3": "WAVE    ",
        "CUNIT3": "angstrom",
        "CDELT3": spectra_pix_size.to_value("angstrom"),
        "CRPIX3": 0,
        "CRVAL3": reference_position[0].to_value("angstrom"),
        "NAXIS3": datacube_shape[0],
    }
    wcs = WCS(wcs_dict)
    return wcs


def build_wcs_from_rss(
    rss_list: list,
    spatial_pix_size: u.Quantity,
    spectra_pix_size: u.Quantity,
    join_type="outer",
    **kwargs,
) -> WCS:
    """Compute the effective WCS resulting from combining an input list of RSS.

    This methods creates a WCS that contains an input list RSS data. The joint
    spectral coverage and field of view, resulting from the combination of the
    individual RSS exposures, can consists of the space that contains all RSS
    (`outer`) or restricted to the region in common among all exposures (`inner`).

    Parameters
    ----------
    rss_list : list or RSS
        List of :class:`RSS`.
    spatial_pix_size : u.Quantity
        Angular spaxel size.
    spectral_pix_size : u.Quantity
        Spectral pixel size.
    join_type : str, optional, default=`outer`
        Which join method to use for creating the combined WCS (`outer` or `inner`).
    **kwargs : dict
        Additional arguments passed to :func:`build_wcs`.

    Returns
    -------
    wcs : astropy.wcs.WCS
        The WCS resulting from combining the list of RSS.
    """
    if isinstance(rss_list, RSS):
        rss_list = [rss_list]

    rss_footprint = np.zeros((len(rss_list), 4, 2)) << u.deg
    rss_spectral_range = np.zeros((len(rss_list), 2)) << u.AA
    for ith, rss in enumerate(rss_list):
        rss_footprint[ith] = rss.get_footprint().to("deg")
        rss_spectral_range[ith] = rss.wavelength[[0, -1]].to("AA")

    # Select the area the cover
    if join_type == "outer":
        max_ra, max_dec = np.nanmax(rss_footprint[:, 0], axis=0)
        min_ra, min_dec = np.nanmin(rss_footprint[:, -1], axis=0)
        min_wl, max_wl = (
            np.nanmin(rss_spectral_range[:, 0]),
            np.nanmax(rss_spectral_range[:, 1]),
        )
    elif join_type == "inner":
        max_ra, max_dec = np.nanmin(rss_footprint[:, 0], axis=0)
        min_ra, min_dec = np.nanmax(rss_footprint[:, -1], axis=0)
        min_wl, max_wl = (
            np.nanmax(rss_spectral_range[:, 0]),
            np.nanmin(rss_spectral_range[:, 1]),
        )

    ra_cen, dec_cen = (max_ra + min_ra) / 2, (max_dec + min_dec) / 2
    ra_width, dec_width = max_ra - min_ra, max_dec - min_dec
    wl_range = max_wl - min_wl

    vprint("Combined footprint center: {:.4f}, {:.4f}".format(ra_cen, dec_cen))
    vprint(
        "Combined footprint Fov: {:.2f}, {:.2f} (arcmin)".format(
            ra_width.to("arcmin"), dec_width.to("arcmin")
        )
    )
    vprint(
        "Combined footprint wavelength range: {:.1f}, {:.1f} (AA)".format(
            min_wl.to("AA"), max_wl.to("AA")
        )
    )

    datacube_shape = (
        int(np.round((wl_range / spectra_pix_size).decompose().value, decimals=0)),
        int(np.round((ra_width / spatial_pix_size).decompose().value, decimals=0)),
        int(np.round((dec_width / spatial_pix_size).decompose().value, decimals=0)),
    )
    reference_position = (min_wl, ra_cen, dec_cen)
    vprint(f"WCS array shape: {datacube_shape} [wave, ra, dec]")
    return build_wcs(
        datacube_shape=datacube_shape,
        reference_position=reference_position,
        spatial_pix_size=spatial_pix_size,
        spectra_pix_size=spectra_pix_size,
        **kwargs,
    )


def make_white_image_from_array(data_array, wavelength=None, **kwargs):
    """Create a white image from a 3D data array.

    Parameters
    ----------
    data_array : np.ndarray
        3D data array. First axis must correspond to the spectral dimension.
    wavelength : np.ndarray
        Wavelength associated to `data_array`.
    **kwargs : dict
        Additional arguments to be passed to :func:`Cube.get_white_image`.

    Return
    ------
    white_image : np.ndarray
        White image array.
    """
    vprint("Creating a Cube from input array")
    cube = Cube(intensity=data_array, wavelength=wavelength)
    return cube.get_white_image(**kwargs)


def make_dummy_cube_from_rss(rss, spa_pix_arcsec=0.5, kernel_pix_arcsec=1.0
                             ) -> Cube:
    """Create an empty datacube array from an input RSS.

    Parameters
    ----------
    rss : :class:`RSS`
        Input RSS
    spa_pix_arcsec : u.Quantity
        Spaxel angular size.
    kernel_pix_arcsec : u.Quantity
        Kernel scale angular size.

    Return
    ------
    cube : :class:`Cube`
    """
    spa_pix_arcsec = ancillary.check_unit(spa_pix_arcsec, u.arcsec)
    kernel_pix_arcsec = ancillary.check_unit(kernel_pix_arcsec, u.arcsec)
    wcs = build_wcs_from_rss(
        rss,
        spatial_pix_size=spa_pix_arcsec,
        spectra_pix_size=rss.wavelength[1] - rss.wavelength[0],
    )
    interpolator = CubeInterpolator(
        [rss],
        pixel_size_arcsec=spa_pix_arcsec,
        wcs=wcs,
        kernel_scale=kernel_pix_arcsec,
    )
    cube = interpolator.build_cube()
    return cube


# =============================================================================
# Mr Krtxo \(ﾟ▽ﾟ)/
#                                                       ... Paranoy@ Rulz! ;^D
